<?php
/**
 * Project: 
 * DZIRcms Content Management System
 * 
 * Description: 
 * Defines the general DZIRcms functions
 * 
 * @package     DZIRcms Backend includes
 * @version     1.0.0
 * @author      René Mansveld
 * @copyright   Spider IT Deutschland <www.spider-it.de>
 * @license     https://www.dzir.org/cms/license/license.txt
 * @link        http://www.spider-it.de
 * @link        https://www.dzir.org/cms
 * @since       file available since drugCMS release 2.0.3
 * 
 * @modified    2015-07-23 René Mansveld :: Added function ScaleImage()
 * @modified    2021-12-07 René Mansveld :: Added function GifToWebp()
 * @modified    2021-12-10 René Mansveld :: Added functionality for resizing animated GIFs
 * 
 */

use GIFEndec\Events\FrameDecodedEvent;
use GIFEndec\IO\FileStream;
use GIFEndec\Decoder;
use GIFEndec\Renderer;
use GIFEndec\Color;
use GIFEndec\Encoder;
use GIFEndec\Frame;

if(!defined('CON_FRAMEWORK')) {
	die('Illegal call');
}

/**
 * convertCmykJpgToSrgbJpg()
 *
 * Converts JPG images using the CMYK color profile to the sRGB color profile
 *
 * @param string $path Full path to the image file
 * @return void
 */
function convertCmykJpgToSrgbJpg($path) {
    if ((strtolower(substr($path, -4)) == '.jpg') || (strtolower(substr($path, -5)) == '.jpeg')) {
        exec('identify -verbose ' . $path . ' >' . $path . '.txt');
        $tmp = file($path . '.txt');
        unlink($path . '.txt');
        for ($i = 0, $n = count($tmp); $i < $n; $i ++) {
            $a = explode(':', $tmp[$i]);
            if (trim($a[0]) == 'Colorspace') {
                if (strpos($a[1], 'RGB') === false) {
                    exec('convert ' . $path . ' -profile sRGB.icc -colorspace sRGB ' . $path . '.jpg');
                    unlink($path);
                    rename($path . '.jpg', $path);
                }
                break;
            }
        }
    }
}

/**
 * isAnimated()
 *
 * Checks if a GIF image is animated
 *
 * @param string $sGifFile Full path to the GIF image file
 * @return bool True if animated, false otherwise
 */
function isAnimated($sGifFile) {
    $filecontents = file_get_contents($sGifFile);
    
    $str_loc = 0;
    $count = 0;
    while ($count < 2) { # There is no point in continuing after we find a 2nd frame
        $where1 = strpos($filecontents, "\x00\x21\xF9\x04", $str_loc);
        if ($where1 === FALSE) {
            break;
        }
        else {
            $str_loc = ($where1 + 1);
            $where2 = strpos($filecontents, "\x00\x2C", $str_loc);
            if ($where2 === FALSE) {
                break;
            }
            else {
                if (($where1 + 8) == $where2) {
                    $count ++;
                }
                $str_loc = ($where2 + 1);
            }
        }
    }
    
    if ($count > 1) {
        return true;
    }
    else {
        return false;
    }
}

/**
 * ScaleImage()
 *
 * Scales and/or zooms an image while keeping it's transparency, saves the new
 * image in the client's cache folder
 *
 * @param string $sImg - Path and file name relative to the client directory
 * @param int $iMaxX - Max width of the new image
 * @param int $iMaxY - Max height of the new image
 * @param bool $bCrop - Image may be cropped (optional)
 * @param bool $bExpand - Image may be expanded (optional)
 * @param float $fGamma - Gamma correction value (1.0 - 1.9) (optional)
 * @param int $iCacheTime - Time setting to overwrite previously generated files (optional)
 * @param bool $bKeepType - New image must be of same type (false generates PNG) (optional)
 * @param bool $bFixedSize - New image must have the exact size with transparent (GIF and PNG) or colored background (optional)
 * @param string $sFixedBG - With $bFixedSize and JPG, this ist the hex color number for the background (optional)
 * @param int $iCropLeft - With $bCrop = true, this is the setting for the left cutting position (-1 = center image) (optional)
 * @param int $iCropTop - With $bCrop = true, this is the setting for the top cutting position (-1 = center image) (optional)
 */
function ScaleImage($sImg, $iMaxX = 0, $iMaxY = 0, $bCrop = false, $bExpand = false, $fGamma = 1.0, $iCacheTime = 10, $bKeepType = false, $bFixedSize = false, $sFixedBG = 'FFFFFF', $iCropLeft = -1, $iCropTop = -1) {
	global $cfgClient, $client, $lang;
    
    $cl = $iCropLeft;
    $ct = $iCropTop;
    
    # Cache
    $md5 = capiImgScaleGetMD5CacheFile($cfgClient[$client]['path']['frontend'] . $sImg, $iMaxX, $iMaxY, $bCrop, $bExpand);
    list($oWidth, $oHeight, $oType) = @getimagesize($cfgClient[$client]['path']['frontend'] . $sImg);
    if (($oType != IMAGETYPE_GIF) && ($oType != IMAGETYPE_JPEG) && ($oType != IMAGETYPE_PNG)) {
        return false;
    }
    if ($bKeepType) {
        switch ($oType) {
            case IMAGETYPE_GIF:
                $cfileName = $md5 . '.gif';
                break;
            case IMAGETYPE_JPEG:
                $cfileName = $md5 . '.jpg';
                break;
            case IMAGETYPE_PNG:
                $cfileName = $md5 . '.png';
                break;
        }
	}
    else {
        $cfileName = $md5 . '.png';
    }
    $cacheDir   = $cfgClient[$client]['path']['frontend'] . 'cache/';
    $cacheFile  = $cacheDir . $cfileName;
	$webFile    = $cfgClient[$client]['path']['htmlpath'] . 'cache/' . $cfileName;
    if (file_exists($cacheFile)) {
        if ($iCacheTime == 0) {
            # File will never be out of date, return it
            return $webFile;
        }
        elseif ((filemtime($cacheFile) + (60 * $iCacheTime)) < time()) {
            # File is out of date
            unlink($cacheFile);
        }
        else {
            # File is still valid, return it
            return $webFile;
        }
    }
    
    # If no size is specified, use the original size
    if ($iMaxX <= 0) {
        $iMaxX = $oWidth;
    }
    if ($iMaxY <= 0) {
        $iMaxY = $oHeight;
    }
    
    # Rebuild the image
    $nLeft      = 0;
    $nTop       = 0;
    $nWidth     = 0;
    $nHeight    = 0;
    $faktor     = 1;
    if ($bFixedSize) {
        $iWidth = $iMaxX;
        $iHeight = $iMaxY;
        # Calculate size and position in the new image
        if (($oWidth > $iMaxX) || ($oHeight > $iMaxY) || ($bExpand)) {
            # Image is larger or must be enhanced
            if ($bCrop) {
                $faktor = max(($iMaxX / $oWidth), ($iMaxY / $oHeight));
            }
            else {
                $faktor = min($iMaxX / $oWidth, $iMaxY / $oHeight);
                $iCropLeft = 0;
                $iCropTop = 0;
            }
            if ($faktor == ($iMaxX / $oWidth)) {
                $nLeft = 0;
                $nWidth = $iMaxX;
                $nHeight = ceil($oHeight * $faktor);
                if ($ct == -1) {
                    $nTop = floor(($iMaxY - $nHeight) / 2);
                }
            }
            else {
                $nTop = 0;
                $nHeight = $iMaxY;
                $nWidth = ceil($oWidth * $faktor);
                if ($cl == -1) {
                    $nLeft = floor(($iMaxX - $nWidth) / 2);
                }
            }
            if ($bCrop) {
                if ($iCropLeft > -1) {
                    $nLeft = 0;
                }
                if ($iCropTop > -1) {
                    $nTop = 0;
                }
            }
        }
        else {
            $nLeft = floor(($iMaxX - $oWidth) / 2);
            $nTop = floor(($iMaxY - $oHeight) / 2);
            $nWidth = $oWidth;
            $nHeight = $oHeight;
        }
    }
    else {
        # Calculate the size of the new image
        if (($oWidth > $iMaxX) || ($oHeight > $iMaxY) || ($bExpand)) {
            if ($bCrop) {
                $faktor = max($iMaxX / $oWidth, $iMaxY / $oHeight);
            }
            else {
                $faktor = min($iMaxX / $oWidth, $iMaxY / $oHeight);
                $iCropLeft = 0;
                $iCropTop = 0;
            }
            if ($faktor == ($iMaxX / $oWidth)) {
                $nWidth = $iMaxX;
                $nHeight = ceil($oHeight * $faktor);
                $iWidth = $iMaxX;
                $iHeight = (($nHeight > $iMaxY) ? $iMaxY : $nHeight);
                if ($ct == -1) {
                    $nTop = (($nHeight > $iMaxY) ? floor(($iMaxY - $nHeight) / 2) : 0);
                }
            }
            else {
                $nHeight = $iMaxY;
                $nWidth = ceil($oWidth * $faktor);
                $iHeight = $iMaxY;
                $iWidth = (($nWidth > $iMaxX) ? $iMaxX : $nWidth);
                if ($cl == -1) {
                    $nLeft = (($nWidth > $iMaxX) ? floor(($iMaxX - $nWidth) / 2) : 0);
                }
            }
            if ($bCrop) {
                if ($iCropLeft > -1) {
                    $nLeft = 0;
                }
                if ($iCropTop > -1) {
                    $nTop = 0;
                }
            }
        }
        else {
            # Image is smaler and must not be enlarged
            $iWidth = $nWidth = $oWidth;
            $iHeight = $nHeight = $oHeight;
        }
    }
    
    if (class_exists('Imagick')) {
        $im = new Imagick($cfgClient[$client]['path']['frontend'] . $sImg);
        $im->resizeImage($nWidth, $nHeight, Imagick::FILTER_CATROM, 0.9, true);
        if ($bCrop) {
            if ($iCropLeft == -1) {
                $iCropLeft = floor(($nWidth - $iWidth) / 2);
            }
            if ($iCropTop == -1) {
                $iCropTop = floor(($nHeight - $iHeight) / 2);
            }
            $im->cropImage($iWidth, $iHeight, abs($iCropLeft), abs($iCropTop));
            $im->setImagePage(0, 0, 0, 0);
        }
        if (($bKeepType == false) && ($oType != IMAGETYPE_PNG)) {
            $im->setImageFormat('png');
        }
        $im->writeImage($cacheFile);
        return $webFile;
    }
    
    if ($iCropLeft < 0) $iCropLeft = 0;
    if ($iCropTop < 0) $iCropTop = 0;
    
    # Read the original image
    switch ($oType) {
        case IMAGETYPE_GIF:
            if (isAnimated($cfgClient[$client]['path']['frontend'] . $sImg)) {
                # Only GIFs are animated (we don't handle WebP here)
                $cachFile = substr($cacheFile, 0, -4) . '.gif';
                $webFile = substr($webFile, 0, -4) . '.gif';
                
                if (!is_dir($cacheDir . 'tmp')) {
                    mkdir($cacheDir . 'tmp', 0755, true);
                }
                
                $tmpGifs = array();
                $_SESSION['cacheDir'] = $cacheDir;
                $_SESSION['gifFile'] = basename($sImg);
                $_SESSION['gifArray'] = $tmpGifs;
                $gifStream = new FileStream($cfgClient[$client]['path']['frontend'] . $sImg);
                $gifDecoder = new Decoder($gifStream);
                $gifRenderer = new Renderer($gifDecoder);
                $gifRenderer->start(function($event) {
                    global $cfgClient, $client;
                    $paddedIndex = str_pad($event->frameIndex, 3, '0', STR_PAD_LEFT);
                    $tmpImg = $_SESSION['cacheDir'] . 'tmp/' . str_replace('.gif', '_frame_' . $paddedIndex . '.gif', $_SESSION['gifFile']);
                    imagegif($event->renderedFrame, $tmpImg);
                    $_SESSION['gifArray'][] = array(
                        'file' => $tmpImg,
                        'disposal_method' => $event->decodedFrame->getDisposalMethod(),
                        'stream' => $event->decodedFrame->getStream(),
                        'duration' => $event->decodedFrame->getDuration(),
                        'transparent_color' => $event->decodedFrame->getTransparentColor()
                    );
                });
                $tmpGifs = $_SESSION['gifArray'];
                $gif = new Encoder();
                foreach ($tmpGifs as $tmpGif) {
                    $tmpImg = $tmpGif['file'];
                    $tImg = ScaleImage(str_replace($cfgClient[$client]['path']['frontend'], '', $tmpImg), $iMaxX, $iMaxY, $bCrop, $bExpand, $fGamma, $iCacheTime, $bKeepType, $bFixedSize, $sFixedBG, $iCropLeft, $iCropTop);
                    if (file_exists(str_replace($cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $tImg))) {
                        unlink($tmpImg);
                        rename(str_replace($cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $tImg), $tmpImg);
                    }
                    $stream = new FileStream($tmpGif['file']);
                    $frame = new Frame();
                    $frame->setDisposalMethod($tmpGif['disposal_method']);
                    $frame->setStream($stream);
                    $frame->setDuration($tmpGif['duration']);
                    if (!is_null($tmpGif['transparent_color'])) {
                        $frame->setTransparentColor($tmpGif['transparent_color']);
                    }
                    else {
                        $frame->setTransparentColor(new Color(255, 255, 255));
                    }
                    $gif->addFrame($frame);
                }
                $gif->addFooter();
                $gif->getStream()->copyContentsToFile($cachFile);
                foreach ($tmpGifs as $tmpGif) {
                    unlink($tmpGif['file']);
                }
                unset($tmpGifs);
                return $webFile;
            }
            else {
                $image = imagecreatefromgif($sImg);
            }
            break;
        case IMAGETYPE_JPEG:
            $image = imagecreatefromjpeg($cfgClient[$client]['path']['frontend'] . $sImg);
            break;
        case IMAGETYPE_PNG:
            $image = imagecreatefrompng($cfgClient[$client]['path']['frontend'] . $sImg);
            break;
        default:
            return false;
    }
    
    # Generate the new image and fill the background
    $nImage = imagecreatetruecolor($iWidth, $iHeight);
    if (($oType == IMAGETYPE_GIF) || ($oType == IMAGETYPE_PNG)) {
        $transIdx = imagecolortransparent($image);
        if ($transIdx >= 0) {
            
            # A transparent color exists (GIF or PNG8)
            $transColor = imagecolorsforindex($image, $transIdx);
            $transIdx = imagecolorallocate($nImage, $transColor['red'], $transColor['green'], $transColor['blue']);
            imagefilledRectangle($nImage, 0, 0, $iWidth, $iHeight, $transIdx);
            imagecolortransparent($nImage, $transIdx);
        } elseif (($oType == IMAGETYPE_PNG) || (!$bKeepType)) {
            
            # PNG24 gets a transparent background per Alpha channel
            $oType = IMAGETYPE_PNG;
            imagealphablending($nImage, false);
            $oColor = imagecolorallocatealpha($nImage, 0, 0, 0, 127);
            imagefilledRectangle($nImage, 0, 0, $iWidth, $iHeight, $oColor);
            imagesavealpha($nImage, true);
        } else {
            
            # Other images get a background color
            $oColor = imagecolorallocate($nImage, hexdec(substr($sFixedBG, 0, 2)), hexdec(substr($sFixedBG, 2, 2)), hexdec(substr($sFixedBG, 4, 2)));
            imagefilledrectangle($nImage, 0, 0, $iWidth, $iHeight, $oColor);
        }
    } else {
        if ($bKeepType) {
            
            # JPG images get a background color
            $oColor = imagecolorallocate($nImage, hexdec(substr($sFixedBG, 0, 2)), hexdec(substr($sFixedBG, 2, 2)), hexdec(substr($sFixedBG, 4, 2)));
            imagefilledrectangle($nImage, 0, 0, $iWidth, $iHeight, $oColor);
        } else {
            
            # JPG images get converted to PNG24
            $oType = IMAGETYPE_PNG;
            imagealphablending($nImage, false);
            $oColor = imagecolorallocatealpha($nImage, 0, 0, 0, 127);
            imagefilledRectangle($nImage, 0, 0, $iWidth, $iHeight, $oColor);
            imagesavealpha($nImage, true);
        }
    }
    
    # Copy the original images scaled into the new one
    ImageCopyResampled($nImage, $image, $nLeft, $nTop, floor($iCropLeft / $faktor), floor($iCropTop / $faktor), ($nWidth + 2), ($nHeight + 2), $oWidth, $oHeight);
    
    # Gamma-correct the new image
    if (($fGamma > 1) && ($fGamma <= 1.9)) {
        imagegammacorrect($nImage, 1, $fGamma);
    }
    
    # Save the new image
    switch ($oType) {
        case IMAGETYPE_GIF:
            imagegif($nImage, $cacheFile);
            break;
        case IMAGETYPE_JPEG:
            imagejpeg($nImage, $cacheFile);
            break;
        case IMAGETYPE_PNG:
            imagepng($nImage, $cacheFile);
            break;
    }
    
    # Cleanup
    imagedestroy($image);
    imagedestroy($nImage);
    
    # Return the new image's path
    return $webFile;
}

function toUint32($n) {
    $ar = unpack("C*", pack("L", $n));
    return $ar;
}
function toUint24($n) {
    $ar = unpack("C*", pack("L", $n));
    array_pop($ar);
    return $ar;
}
function toUint16($n) {
    $ar = unpack("C*", pack("S", $n));
    return $ar;
}

function bytesToString($bytes) {
    return implode(array_map("chr", $bytes));
}

function binaryToBytes($bits) {
    $octets = explode(' ', $bits);
    return array_map("bindec", $octets);
}

/**
 * _getFrameData()
 *
 * Helper function for GifToWebp()
 */
function _getFrameData($image, $msec) {
    $w = imagesx($image);
    $h = imagesy($image);

    ob_start();
    imagewebp($image);
    if ((ob_get_length() % 2) == 1) {
        echo "\0";
    }
    $image_data = ob_get_contents();
    ob_end_clean();
    $frameData = substr($image_data, strpos($image_data, 'VP8 '));
    return Array(
        'frameData' => $frameData,
        'duration' => bytesToString(toUint24($msec)),
        'width' => bytesToString(toUint24($w - 1)),
        'height' => bytesToString(toUint24($h -1 )),
    );
}

/**
 * GifToWebp()
 *
 * Converts GIF images to WebP while keeping the animation
 *
 * @param string $sImg Path and file name relative to the client directory
 * @param int $iCacheTime - Time setting to overwrite previously generated files (optional)
 * @return string Path and file name relative to the client directory
 */
function GifToWebp($sImg, $iCacheTime = 0) {
	global $cfgClient, $client, $lang;
    
    list($oWidth, $oHeight, $oType) = @getimagesize($cfgClient[$client]['path']['frontend'] . $sImg);
    if ($oType != IMAGETYPE_GIF) {
        return false;
    }
    $sFileName  = str_replace('.gif', '.webp', basename($sImg));
    $cacheDir   = $cfgClient[$client]['path']['frontend'] . 'cache/';
    $cacheFile  = $cacheDir . $sFileName;
	$webFile    = 'cache/' . $sFileName;
    if (file_exists($cacheFile)) {
        if ($iCacheTime == 0) {
            # File will never be out of date, return it
            return $webFile;
        }
        elseif ((filemtime($cacheFile) + (60 * $iCacheTime)) < time()) {
            # File is out of date
            unlink($cacheFile);
        }
        else {
            # File is still valid, return it
            return $webFile;
        }
    }
    
    if (isAnimated($cfgClient[$client]['path']['frontend'] . $sImg)) {
        $_SESSION['gifArray'] = array();
        $gifStream = new FileStream($cfgClient[$client]['path']['frontend'] . $sImg);
        $gifDecoder = new Decoder($gifStream);
        $gifRenderer = new Renderer($gifDecoder);
        $gifRenderer->start(function($event) {
            $_SESSION['gifArray'][] = _getFrameData($event->renderedFrame, $event->decodedFrame->getDuration());
        });
        $frameArray = $_SESSION['gifArray'];
        unset($_SESSION['gifArray']);
        
        # Create new WebP image
        $fileWEBP = '';
        $fileHeader = '';
        $fileContents = '';
        
        # Chunk Header VP8X
        $fileContents .= 'VP8X';
        $headChunkSize = bytesToString(toUint32(10));
        # bit flags Rsv|I|L|E|X|A|R|                   Reserved
        $oVP8XflagsBin = '00010010 00000000 00000000 00000000';
        $oVP8Xflags = bytesToString(binaryToBytes($oVP8XflagsBin));
        $oCanvasSize = bytesToString(toUint24($oWidth-1)) . bytesToString(toUint24($oHeight-1));
        $fileContents .= $headChunkSize . $oVP8Xflags . $oCanvasSize;
        
        # Chunk Header ANIM
        $fileContents .= 'ANIM';
        $animChunkSize = bytesToString(toUint32(6));
        # loop count 16bits, 0 = infinito
        # bytesToString(toUint16(0));
        $oLoopCount = str_repeat(chr(0), 2);
        # 32bits BGRA, Blue Green Red Alpha (0, 0, 0, 0)
        $oBackGround = str_repeat(chr(0), 4);
        $fileContents .= $animChunkSize . $oBackGround . $oLoopCount;
        
        foreach ($frameArray as $frame) {
            # Chunk HEADER ANMF
            $fileContents .= 'ANMF';
            $frameDataChunkSize = bytesToString(toUint32(strlen($frame['frameData']) + 16));
            # frame origin X Y
            # bytesToString(toUint24(originX)) . bytesToString(toUint24(originY))
            # (0, 0)
            $fOrigin = str_repeat(chr(0), 6);
            # frame size (uint24) width - 1 , (uint24) height - 1
            $fSize = $frame['width'] . $frame['height'];
            # frame duration in miliseconds (uint24)
            $fDuration = $frame['duration'];
            # frame options bits
            # reserved (6 bits) + alpha blending (1 bit) + descartar frame (1 bit)
            $fFlagsBin = '00000010';
            $fFlags = bytesToString(binaryToBytes($fFlagsBin));
            # chunk payload
            $fileContents .= $frameDataChunkSize . $fOrigin . $fSize . $fDuration . $fFlags . $frame['frameData'];
        }
        
        # calculate Size and build file header
        $fileSize = bytesToString(toUint32(strlen($fileContents) + 4));
        $fileHeader = 'RIFF' . $fileSize . 'WEBP';
        $fileWEBP = $fileHeader . $fileContents;
        
        file_put_contents($cacheFile, $fileWEBP);
    }
    else {
        $im = imagecreatefromgif($cfgClient[$client]['path']['frontend'] . $sImg);
        $result = imagewebp($im, $cacheFile);
        if ($result === false) {
            $webFile = substr($webFile, 0, -5) . substr($sImg, strrpos($sImg, '.'));
        }
        imagedestroy($im);
    }
    return $webFile;
}

?>