<?php
/**
 * Project: 
 * DZIRcms Content Management System
 * 
 * Description: 
 * Generate compressed output for frontend css and js files
 * 
 * Client setting to switch output compression must look like this:
 * Type:    generator
 * Name:    output_compression
 * Value:   full / correct / off
 *          full = compress css and js files, link them in the head/body area (default)
 *          correct = move css and javascript file links to the head/body area
 *          off = no correction and compression (use for development only)
 * Value:   fonts
 *          Added to the other value, it downloads remote fonts and delivers them locally
 * 
 * Requirements: 
 * @con_php_req 5.2
 * 
 *
 * @package    DZIRcms Frontend classes
 * @version    1.0.1
 * @author     René Mansveld
 * @copyright  Spider IT Deutschland <www.spider-it.de>
 * @license    https://www.dzir.org/cms/license/LIZENZ.txt
 * @link       https://www.dzir.org/cms
 * 
 * {@internal 
 *   created:   2012-12-16
 *   modified:  2018-09-25 René Mansveld    Added code to download and deliver remote fonts
 *   modified:  2019-01-30 René Mansveld    Added code to compile LESS files (Note: not when switched off)
 *   modified:  2019-02-01 René Mansveld    Added code to compile SCSS files (Note: not when switched off)
 *   modified:  2020-02-07 René Mansveld    Corrected recognition of URLs starting with '//' as external URLs
 *   modified:  2020-12-14 René Mansveld    Added code to convert images to WebP for smaller files
 *   modified:  2020-12-15 René Mansveld    Added code to move <link rel="preload"> entries after CSS links
 *   modified:  2021-03-19 René Mansveld    Added code to convert images INSIDE CSS to WebP for smaller files
 *   modified:  2021-05-27 René Mansveld    Code correction to not add the url domain part to url(data:...)
 *   modified:  2021-06-23 René Mansveld    Added code to replace <DZIRcms> tags to HTML comments for the header code image
 *   modified:  2021-09-21 René Mansveld    Optimized code to compile LESS and SCSS files even when switched off
 *   
 * }}
 * 
 */

if(!defined('CON_FRAMEWORK')) {
    die('Illegal call');
}

/**
 * Helper-Function _convertImagesFromCodeToWebP
 * 
 * Searches the code of a file for images, converts the mentioned
 * images to WebP, and then exchanges them in the code.
 * 
 * @param string $sCode - The code to rework
 * @param string $sFilePath - The path of the code containing file
 * @param int $iNumSubdirs - The number of subdirs from the
 *        client's directory for the new file
 * @return string - The reworked code
 */
function _convertImagesFromCodeToWebP($sCode, $sFilePath, $iNumSubdirs, $bDebug = false) {
    global $cfgClient, $client;
    
    if (strpos($_SERVER['HTTP_ACCEPT'], 'image/webp') !== false) {
        $sFilePath = ((substr($sFilePath, -1, 1) == '/') ? substr($sFilePath, 0, -1) : $sFilePath);
        
        # Convert PNG images to WebP if the browser supports this format
        $p1 = strpos(strtolower($sCode), '.png');
        while ($p1 !== false) {
            $p1 += 4;
            switch (substr($sCode, $p1, 1)) {
                case '"':
                    $p2 = (strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case "'":
                    $p2 = (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case ')':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, '(', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
                case ' ':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ' ', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
            }
            if (substr(substr($sCode, 0, $p2), -7, 6) == 'value=') {
                $p1 = strpos(strtolower($sCode), '.jpg', $p1);
                continue;
            }
            $orig = substr($sCode, $p2, ($p1 - $p2));
            $tmp = $orig;
            if (substr($tmp, 0, 4) == 'http') {
                
                # Remove the webpath from the fullpath if it is the client's webpath
                $tmp = str_replace($cfgClient[$client]['path']['htmlpath'], '', $tmp);
            }
            if (substr($tmp, 0, 4) == 'http') {
                
                # Leave image paths to other domains as they are
                $p1 = strpos(strtolower($sCode), '.png', $p1);
            }
            else {
                
                # Generate the fullpath for the WebP image
                $webp = 'cache/' . str_replace('.png', '.webp', basename($tmp));
                
                # Check if this file already exists, generate it when not
                if (!file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    imagewebp(imagecreatefrompng(realpath($sFilePath . '/' . $tmp)), $cfgClient[$client]['path']['frontend'] . $webp);
                }
                
                # Check again if the file exists and replace it in the code if it does
                if (file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    $new = str_repeat('../', $iNumSubdirs);
                    $sCode = substr($sCode, 0, $p2) . $new . $webp . substr($sCode, $p1);
                    $p1 = strpos(strtolower($sCode), '.png', $p2);
                }
                else {
                    $p1 = strpos(strtolower($sCode), '.png', $p1);
                }
            }
        }
        
        # Convert JPG images to WebP if the browser supports this format
        $p1 = strpos(strtolower($sCode), '.jpg');
        while ($p1 !== false) {
            $p1 += 4;
            switch (substr($sCode, $p1, 1)) {
                case '"':
                    $p2 = (strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case "'":
                    $p2 = (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case '(':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ')', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
                case ' ':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ' ', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
            }
            if (substr(substr($sCode, 0, $p2), -7, 6) == 'value=') {
                $p1 = strpos(strtolower($sCode), '.jpg', $p1);
                continue;
            }
            $orig = substr($sCode, $p2, ($p1 - $p2));
            $tmp = $orig;
            if (substr($tmp, 0, 4) == 'http') {
                
                # Remove the webpath from the fullpath if it is the client's webpath
                $tmp = str_replace($cfgClient[$client]['path']['htmlpath'], '', $tmp);
            }
            if (substr($tmp, 0, 4) == 'http') {
                
                # Leave image paths to other domains as they are
                $p1 = strpos(strtolower($sCode), '.jpg', $p1);
            }
            else {
                
                # Generate the fullpath for the WebP image
                $webp = 'cache/' . str_replace('.jpg', '.webp', basename($tmp));
                
                # Check if this file already exists, generate it when not
                if (!file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    imagewebp(imagecreatefromjpeg(realpath($sFilePath . '/' . $tmp)), $cfgClient[$client]['path']['frontend'] . $webp);
                }
                
                # Check again if the file exists and replace it in the code if it does
                if (file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    $new = str_repeat('../', $iNumSubdirs);
                    $sCode = substr($sCode, 0, $p2) . $new . $webp . substr($sCode, $p1);
                    $p1 = strpos(strtolower($sCode), '.jpg', $p2);
                }
                else {
                    $p1 = strpos(strtolower($sCode), '.jpg', $p1);
                }
            }
        }
        $p1 = strpos(strtolower($sCode), '.jpeg');
        while ($p1 !== false) {
            $p1 += 5;
            switch (substr($sCode, $p1, 1)) {
                case '"':
                    $p2 = (strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case "'":
                    $p2 = (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case '(':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ')', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
                case ' ':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ' ', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
            }
            if (substr(substr($sCode, 0, $p2), -7, 6) == 'value=') {
                $p1 = strpos(strtolower($sCode), '.jpg', $p1);
                continue;
            }
            $orig = substr($sCode, $p2, ($p1 - $p2));
            $tmp = $orig;
            if (substr($tmp, 0, 4) == 'http') {
                
                # Remove the webpath from the fullpath if it is the client's webpath
                $tmp = str_replace($cfgClient[$client]['path']['htmlpath'], '', $tmp);
            }
            if (substr($tmp, 0, 4) == 'http') {
                
                # Leave image paths to other domains as they are
                $p1 = strpos(strtolower($sCode), '.jpeg', $p1);
            }
            else {
                
                # Generate the fullpath for the WebP image
                $webp = 'cache/' . str_replace('.jpg', '.webp', basename($tmp));
                
                # Check if this file already exists, generate it when not
                if (!file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    imagewebp(imagecreatefromjpeg(realpath($sFilePath . '/' . $tmp)), $cfgClient[$client]['path']['frontend'] . $webp);
                }
                
                # Check again if the file exists and replace it in the code if it does
                if (file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    $new = str_repeat('../', $iNumSubdirs);
                    $sCode = substr($sCode, 0, $p2) . $new . $webp . substr($sCode, $p1);
                    $p1 = strpos(strtolower($sCode), '.jpeg', $p2);
                }
                else {
                    $p1 = strpos(strtolower($sCode), '.jpeg', $p1);
                }
            }
        }
        
        # Convert GIF images to WebP if the browser supports this format
        $p1 = strpos(strtolower($sCode), '.gif');
        while ($p1 !== false) {
            $p1 += 4;
            switch (substr($sCode, $p1, 1)) {
                case '"':
                    $p2 = (strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case "'":
                    $p2 = (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1);
                    break;
                case '(':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ')', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
                case ' ':
                    $p2 = max(array((strrpos($sCode, '"', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, "'", -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ' ', -(strlen($sCode) - $p1 + 1)) + 1), (strrpos($sCode, ',', -(strlen($sCode) - $p1 + 1)) + 1)));
                    break;
            }
            if (substr(substr($sCode, 0, $p2), -7, 6) == 'value=') {
                $p1 = strpos(strtolower($sCode), '.jpg', $p1);
                continue;
            }
            $orig = substr($sCode, $p2, ($p1 - $p2));
            $tmp = $orig;
            if (substr($tmp, 0, 4) == 'http') {
                
                # Remove the webpath from the fullpath if it is the client's webpath
                $tmp = str_replace($cfgClient[$client]['path']['htmlpath'], '', $tmp);
            }
            if (substr($tmp, 0, 4) == 'http') {
                
                # Leave image paths to other domains as they are
                $p1 = strpos(strtolower($sCode), '.gif', $p1);
            }
            else {
                
                # Generate the fullpath for the WebP image
                $webp = 'cache/' . str_replace('.jpg', '.webp', basename($tmp));
                
                # Check if this file already exists, generate it when not
                if (!file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    GifToWebp($tmp);
                }
                
                # Check again if the file exists and replace it in the code if it does
                if (file_exists($cfgClient[$client]['path']['frontend'] . $webp)) {
                    $new = str_repeat('../', $iNumSubdirs);
                    $sCode = substr($sCode, 0, $p2) . $new . $webp . substr($sCode, $p1);
                    $p1 = strpos(strtolower($sCode), '.gif', $p2);
                }
                else {
                    $p1 = strpos(strtolower($sCode), '.gif', $p1);
                }
            }
        }
        
        return $sCode;
    }
}

/**
 * Helper-Function _correctLinkedPaths
 * 
 * Searches the code of a file for linked files and corrects their
 * paths to the new saved position.
 * 
 * @param string $sCode - The code to rework
 * @param string $sOldFilePath - The path of the original file
 * @param int $iNumSubdirs - The number of subdirs from the
 *        client's directory for the new file
 * @return string - The reworked code
 */
function _correctLinkedPaths($sCode, $sOldFilePath, $iNumSubdirs) {
    global $cfgClient, $client;
    
    $sOldRelPath = str_replace($cfgClient[$client]['path']['frontend'], '', $sOldFilePath);
    $iNumOldDirs = count(explode('/', substr($sOldRelPath, 0, -1)));
    $p1 = strpos($sCode, 'url(');
    while ($p1 !== false) {
        $p1 += 4;
        $c = substr($sCode, $p1, 1);
        if (in_array($c, array('"', "'"))) {
            $p1 += 1;
            $p2 = strpos($sCode, $c, $p1);
        }
        else {
            $p2 = strpos($sCode, ')', $p1);
        }
        $to = substr($sCode, $p1, ($p2 - $p1));
        if ((substr($to, 0, 8) == 'https://') || (substr($to, 0, 7) == 'http://') || (substr($to, 0, 2) == '//')) {
            # Absolute path, correct it to a relative path if the target file is in the client's directory
            if (substr($to, 0, strlen($cfgClient[$client]['path']['htmlpath'])) == $cfgClient[$client]['path']['htmlpath']) {
                $tn = str_repeat('../', $iNumOldDirs) . str_replace($cfgClient[$client]['path']['htmlpath'], '', $to);
            }
            else {
                $tn = $to;
            }
        }
        else {
            $tn = $to;
        }
        if (substr($tn, 0, 5) == 'data:') {
            # Plain data, leave it alone
        }
        elseif ((substr($tn, 0, 8) == 'https://') || (substr($tn, 0, 7) == 'http://') || (substr($tn, 0, 2) == '//')) {
            # Absolute path, leave it alone
        }
        else {
            $tn = realpath($sOldFilePath . '/' . $tn);
            $tn = str_repeat('../', $iNumSubdirs) . str_replace(array($cfgClient[$client]['path']['frontend'], '../'), '', $to);
            $sCode = substr($sCode, 0, $p1) . $tn . substr($sCode, $p2);
        }
        $p1 = strpos($sCode, 'url(', $p1);
    }
    return $sCode;
}

/**
 * Chain-Function cecOutputCompressor
 * 
 * Checks the HTML code for CSS and JS files, hands those to the
 * OutputCompressor class to join and compact them, replaces them
 * in the HTML code in the <head> / <body> section, and then
 * cleans and compacts the HTML code.
 * 
 * @param string $sCode - The HTML code of the page
 * @return string - The new HTML code of the page
 */
function cecOutputCompressor($sCode) {
    global $cfgClient, $client, $edit;
    
    $sCacheDir = $cfgClient[$client]['path']['frontend'] . 'cache/';
    
    $sDbgCode = '';
    
    # How is output compression switched?
    $switch = getEffectiveSetting('generator', 'output_compression', 'full', 'Correct or compress the output', 'With this setting you control the output compressor and tell it wether to compress, download fonts, and convert images', array('off', 'correct', 'correct,fonts', 'correct,images', 'correct,fonts,images', 'full', 'full,fonts', 'full,images', 'full,fonts,images'));
    if ($edit) {
        # In editor, only generate CSS out of LESS/SCSS
        $switch = 'off';
    }
    if ($switch == 'off') {
        if (!is_dir($sCacheDir . 'css')) {
            mkdir($sCacheDir . 'css', 0755, true);
        }
        
        # Find all link tags with rel="stylesheet"
        $stylesheets = array();
        $iStyles = 0;
        $p1 = strpos($sCode, '<link');
        while ($p1 !== false) {
            $p2 = strpos($sCode, '>', $p1);
            $tmp = substr($sCode, $p1, (($p2 - $p1) + 1));
            if ((strpos(strtolower($tmp), 'rel="stylesheet"')) || (strpos(strtolower($tmp), "rel='stylesheet'"))) {
                # Extract the path and filename
                $url = '';
                $p3 = strpos($tmp, 'href=');
                if ($p3 !== false) { # shouldn't happen, an address is needed
                    $p3 += 6;
                    $char = substr($tmp, ($p3 - 1), 1);
                    $p4 = strpos($tmp, $char, $p3);
                    $url = substr($tmp, $p3, ($p4 - $p3));
                    # Clean the url from unnessecary path information (just leave the folder name like "css/xxx.css")
                    $url = str_replace($cfgClient[$client]['path']['htmlpath'], '', $url);
                    $url = (((substr($url, 0, 1) == '/') && (substr($url, 0, 2) != '//')) ? substr($url, 1) : $url);
                    # Exclude files with complete URLs or parameters, and ordinary css files
                    $compress = (((substr($url, 0, 7) == 'http://') || (substr($url, 0, 8) == 'https://') || (substr($url, 0, 2) == '//') || (strpos($url, '?') !== false) || (substr($url, -4) == '.css')) ? false : true);
                    if ($compress) {
                        if ((substr($url, -5) == '.less') && (is_file($cfgClient[$client]['path']['frontend'] . $url))) {
                            if (!is_dir($sCacheDir . 'less_compiled')) {
                                mkdir($sCacheDir . 'less_compiled', 0755, true);
                            }
                            $sFileName = 'cache/less_compiled/' . Less_Cache::Get(array($cfgClient[$client]['path']['frontend'] . $url => dirname($url)), array('cache_dir' => $sCacheDir . 'less_compiled/'));
                            rename($cfgClient[$client]['path']['frontend'] . $sFileName, $cfgClient[$client]['path']['frontend'] . $sFileName . '.' . filemtime($cfgClient[$client]['path']['frontend'] . $url) . '.css');
                            $sFileName .= '.' . filemtime($cfgClient[$client]['path']['frontend'] . $url) . '.css';
                            $c = file_get_contents($cfgClient[$client]['path']['frontend'] . $sFileName);
                            $c = str_replace(array("url('", "url('../../"), "url('../", $c);
                            $c = str_replace(array($cfgClient[$client]['path']['htmlpath'] . 'data:', $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:'), 'data:', $c);
                            file_put_contents($cfgClient[$client]['path']['frontend'] . $sFileName, $c);
                        }
                        elseif ((substr($url, -5) == '.scss') && (is_file($cfgClient[$client]['path']['frontend'] . $url))) {
                            if (!is_dir($sCacheDir . 'scss_compiled')) {
                                mkdir($sCacheDir . 'scss_compiled', 0755, true);
                            }
                            # Check if a compiled file with the same name exists in the cache dir, and if so, if it is newer than this file
                            $sFN = basename($url);
                            $sFN = substr($sFN, 0, strrpos($sFN, '.')) . '.css.' . filemtime($cfgClient[$client]['path']['frontend'] . $url) . '.css';
                            if ((file_exists($sCacheDir . 'scss_compiled/' . $sFN)) && (filemtime($sCacheDir . 'scss_compiled/' . $sFN) >= filemtime($cfgClient[$client]['path']['frontend'] . $url)) && (filesize($sCacheDir . 'scss_compiled/' . $sFN) > 0)) {
                                $sFileName = 'cache/scss_compiled/' . $sFN;
                            }
                            else {
                                cInclude('external', 'Scss/Base/Range.php');
                                cInclude('external', 'Scss/Block.php');
                                cInclude('external', 'Scss/Colors.php');
                                cInclude('external', 'Scss/Compiler.php');
                                cInclude('external', 'Scss/Compiler/Environment.php');
                                cInclude('external', 'Scss/Exception/CompilerException.php');
                                cInclude('external', 'Scss/Exception/ParserException.php');
                                cInclude('external', 'Scss/Exception/RangeException.php');
                                cInclude('external', 'Scss/Exception/ServerException.php');
                                cInclude('external', 'Scss/Formatter.php');
                                cInclude('external', 'Scss/Formatter/Compact.php');
                                cInclude('external', 'Scss/Formatter/Compressed.php');
                                cInclude('external', 'Scss/Formatter/Crunched.php');
                                cInclude('external', 'Scss/Formatter/Debug.php');
                                cInclude('external', 'Scss/Formatter/Expanded.php');
                                cInclude('external', 'Scss/Formatter/Nested.php');
                                cInclude('external', 'Scss/Formatter/OutputBlock.php');
                                cInclude('external', 'Scss/Node.php');
                                cInclude('external', 'Scss/Node/Number.php');
                                cInclude('external', 'Scss/Parser.php');
                                cInclude('external', 'Scss/SourceMap/Base64.php');
                                cInclude('external', 'Scss/SourceMap/Base64VLQ.php');
                                cInclude('external', 'Scss/SourceMap/SourceMapGenerator.php');
                                cInclude('external', 'Scss/Type.php');
                                cInclude('external', 'Scss/Util.php');
                                cInclude('external', 'Scss/Version.php');
                                try {
                                    $scss = new Leafo\ScssPhp\Compiler();
                                    $scss->setFormatter('Leafo\ScssPhp\Formatter\Expanded');
                                    $scss->setImportPaths($cfgClient[$client]['path']['frontend'] . dirname($url) . '/');
                                    $sCompiled = $scss->compile(file_get_contents($cfgClient[$client]['path']['frontend'] . $url));
                                    $sCompiled = str_replace(array($cfgClient[$client]['path']['htmlpath'] . 'data:', $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'), 'data:', $sCompiled);
                                    if (file_put_contents($cfgClient[$client]['path']['frontend'] . 'cache/scss_compiled/' . $sFN, $sCompiled) !== false) {
                                        $sFileName = 'cache/scss_compiled/' . $sFN;
                                    }
                                    else {
                                        $sFN = 'cache/css/' . basename($url) . filemtime($cfgClient[$client]['path']['frontend'] . $url) . 'css';
                                        copy($cfgClient[$client]['path']['frontend'] . $url, $cfgClient[$client]['path']['frontend'] . $sFN);
                                        $sFileName = $sFN;
                                    }
                                    unset($scss);
                                } catch (\Exception $e) {
                                    $refl = new ReflectionClass($e);
                                    $prop = $refl->getProperty('message');
                                    $prop->setAccessible(true);
                                    $p1 = strpos($sCode, '<body');
                                    $p2 = (strpos($sCode, '>', $p1) + 1);
                                    $sCode = substr($sCode, 0, $p2) . "\n" . '<h6>SCSS error: ' . $prop->getValue($e) . '</h6>' . substr($sCode, $p2);
                                    $sFN = 'cache/css/' . basename($url) . filemtime($cfgClient[$client]['path']['frontend'] . $url) . 'css';
                                    copy($cfgClient[$client]['path']['frontend'] . $url, $cfgClient[$client]['path']['frontend'] . $sFN);
                                    touch($cfgClient[$client]['path']['frontend'] . $sFN, filemtime($cfgClient[$client]['path']['frontend'] . $url));
                                    $sFileName = $sFN;
                                }
                            }
                        }
                        $sCode = str_replace($url, $sFileName, $sCode);
                    }
                }
            }
            $p1 = strpos($sCode, '<link', $p2);
        }
        
        
        return str_replace(array('<DZIRcms>', '</DZIRcms>'), array('<!--', '-->'), $sCode);
    }
    
    # Remove comments (NOT IE conditional comment tags and HTML comments in Javascript blocks) from the code
    $p0 = strpos($sCode, '</body>');
    $p1 = strpos($sCode, '<!--');
    while (($p1 !== false) && ($p1 < $p0)) {
        if (substr($sCode, $p1, 7) == '<!--[if') {
            # IE conditional comment start tag
            $p2 = strpos($sCode, ']-->', ($p1 + 1));
            if (($p2 !== false) && ($p2 > $p1)) {
                $p1 = strpos($sCode, '<!--', ($p2 + 1));
            } else {
                $p1 = strpos($sCode, '<!--', ($p1 + 7));
            }
        } elseif ((substr(str_replace(array("\n", ' '), '', substr($sCode, 0, $p1)), -30) == '<script type="text/javascript">') || (substr(str_replace(array("\n", ' '), '', substr($sCode, 0, $p1)), -30) == "<script type='text/javascript'>")) {
            # HTML comment in Javascript block (e.g. <script type="text/javascript"><!--  --></script>)
            $p1 = strpos($sCode, '<!--', ($p1 + 1));
        } else {
            $p2 = strpos($sCode, '-->', $p1);
            while (substr($sCode, ($p2 - 6), 9) == 'endif]-->') {
                # IE conditional comment end tag, but in commented block
                $p2 = strpos($sCode, '-->', ($p2 + 1));
            }
            $tmp = substr($sCode, $p1, ($p2 - $p1 + 3));
            $sCode = str_replace($tmp, '', $sCode);
            $p0 = strpos($sCode, '</body>');
            $p1 = strpos($sCode, '<!--', $p1);
        }
    }
    
    # If switched on, find links to external fonts, copy the font files to the own webspace, and change the links to local ones
    if (strpos($switch, 'fonts') !== false) {
        $sPath = $sCacheDir . 'fonts/';
        
        # Find all embedded fonts by their tags
        $cnt = 0;
        $p1 = strpos($sCode, '<link ');
        while ($p1 !== false) {
            $p2 = strpos($sCode, '>', $p1);
            $tmp = substr($sCode, $p1, ($p2 - $p1));
            if ((strpos($tmp, 'rel="stylesheet"') !== false) || (strpos($tmp, "rel='stylesheet'") !== false) || (strpos($tmp, 'type="text/css"') !== false) || (strpos($tmp, "type='text/css'") !== false)) {
                $p3 = strpos($tmp, 'href=');
                if ($p3 !== false) {
                    $char = substr($tmp, ($p3 + 5), 1);
                    $p4 = strpos($tmp, $char, ($p3 + 6));
                    $url = substr($tmp, ($p3 + 6), ($p4 - $p3 - 6));
                    if ((substr($url, 0, 7) == 'http://') || (substr($url, 0, 8) == 'https://') || (substr($url, 0, 2) == '//')) {
                        $cnt ++;
                        if (!is_dir($sPath)) {
                            @mkdir($sPath, 0755, true);
                        }
                        $errno = 0;
                        $errmsg = '';
                        getRemoteContentToFile(((substr($url, 0, 2) == '//') ? 'https:' : '') . $url, $sPath . 'temp.css', $errno, $errmsg);
                        
                        # Get the name of the font family
                        $cnt = file_get_contents($sPath . 'temp.css');
                        $p5 = strpos($cnt, 'font-family');
                        if ($p5 !== false) {
                            $p6 = strpos($cnt, ';', $p5);
                            $name = trim(str_replace(array('"', "'", ' '), '', substr($cnt, ($p5 + 12), ($p6 - $p5 - 12))));
                        }
                        else {
                            $name = 'font' . substr('0000' . $cnt, -5);
                        }
                        
                        # Rename the font definition file if the previously downloaded version is at least one day old
                        if ((!file_exists($sPath . $name . '.css')) || (filemtime($sPath . $name . '.css') < strtotime('-1 day'))) {
                            if (file_exists($sPath . $name . '.css')) {
                                unlink($sPath . $name . '.css');
                            }
                            rename($sPath . 'temp.css', $sPath . $name . '.css');
                            
                            # Now (re)download all font files in the font definition file
                            $p5 = strpos($cnt, 'url(');
                            while ($p5 !== false) {
                                $p6 = strpos($cnt, ')', $p5);
                                $lnk = trim(str_replace(array('"', "'"), '', substr($cnt, ($p5 + 4), ($p6 - $p5 - 4))));
                                $errno = 0;
                                $errmsg = '';
                                getRemoteContentToFile(((substr($lnk, 0, 2) == '//') ? 'https:' : '') . $lnk, $sPath . basename($lnk), $errno, $errmsg);
                                $p5 = strpos($cnt, 'url(', $p6);
                            }
                            
                            # Now exchange the urls in the font definition file
                            $cnt = str_replace(dirname($lnk), 'cache/fonts', $cnt);
                            
                            # Overwite the previously renamed font definition file
                            $fp = fopen($sPath . $name . '.css', 'w');
                            fputs($fp, $cnt);
                            fclose($fp);
                        }
                        else {
                            unlink($sPath . 'temp.css');
                        }
                        
                        # Exchange the link to the remote font definition file with a link to our local one
                        $sCode = substr($sCode, 0, $p1) . str_replace($url, 'cache/fonts/' . $name . '.css', $tmp) . substr($sCode, $p2);
                    }
                }
            }
            $p1 = strpos($sCode, '<link ', ($p1 + 1));
        }
    }
    
    # First find all style and script link tags and script blocks inside IE conditional comments (eg <!--[if IE]>...<![endif]-->)
    # We can move the complete block to the end of the head/body section if only link tags are included, otherwise we must move only
    # the links while rebuilding the IE conditional comments around them
    $IE = array();
    $IElinks = array();
    $IEscripts = array();
    $IEscriptblocks = array();
    $IEstyles = array();
    $p1 = strpos($sCode, '<!--[if');
    while ($p1 !== false) {
        # First get the exact start comment
        $p2 = strpos($sCode, '>', $p1);
        $c = substr($sCode, $p1, (($p2 - $p1) + 1));
        if (substr($sCode, ($p2 + 1), 5) == '<!-->') {
            $c .= '<!-->';
        }
        $IEtmp = array();
        $p2 = strpos($sCode, 'endif]-->', $p1);
        if (($p2 !== false) && ($p2 > $p1)) {
            $tmp = substr($sCode, $p1, (($p2 - $p1) + 9));
            # Find all style link tags with rel="stylesheet"
            $p3 = strpos($tmp, '<link');
            while ($p3 !== false) {
                $p4 = strpos($tmp, '>', $p3);
                $tmp2 = substr($tmp, $p3, (($p4 - $p3) + 1));
                if ((strpos(strtolower($tmp2), 'rel="stylesheet"')) || (strpos(strtolower($tmp2), "rel='stylesheet'"))) {
                    # Extract the path and filename
                    $url = '';
                    $p5 = strpos($tmp2, 'href=');
                    if ($p5 !== false) { # shouldn't happen, an address is needed
                        $p5 += 6;
                        $char = substr($tmp2, ($p5 - 1), 1);
                        $p6 = strpos($tmp2, $char, $p5);
                        $url = substr($tmp2, $p5, ($p6 - $p5));
                        # Clean the url from unnessecary path information (just leave the folder name like "css/xxx.css")
                        $url = str_replace($cfgClient[$client]['path']['htmlpath'], '', $url);
                        $url = (((substr($url, 0, 1) == '/') && (substr($url, 0, 2) != '//')) ? substr($url, 1) : $url);
                        # Sort into array grouped by media type
                        $p5 = strpos($tmp2, 'media=');
                        if ($p5 !== false) {
                            $p5 += 7;
                            $char = substr($tmp2, ($p5 - 1), 1);
                            $p6 = strpos($tmp2, $char, $p5);
                            $media = substr($tmp2, $p5, ($p6 - $p5));
                            $IElinks[$c][$media][] = array('url' => $url, 'old' => $tmp2);
                        } else {
                            $IElinks[$c]['all'][] = array('url' => $url, 'old' => $tmp2);
                        }
                        $IEtmp[] = $tmp2;
                    }
                }
                $p3 = strpos($tmp, '<link', $p4);
            }
            # Find all script link tags and script blocks
            $p3 = strpos($tmp, '<script');
            while ($p3 !== false) {
                $p4 = strpos($tmp, '>', $p3);
                # Check if this is included code, or code directly in the page
                if (substr($tmp, $p4, 10) == '></script>') {
                    $tmp2 = substr($tmp, $p3, (($p4 - $p3) + 10));
                    # Check if this script tag has a type or language set
                    $p5 = strpos($tmp2, 'type=');
                    if ($p5 !== false) {
                        $p5 += 6;
                        $char = substr($tmp2, ($p5 - 1), 1);
                        $p6 = strpos($tmp2, $char, $p5);
                        $type = substr($tmp2, $p5, ($p6 - $p5));
                        # If type isn't "text/javascript", leave the tag alone
                        if (strtolower($type) != 'text/javascript') {
                            continue;
                        }
                    } else {
                        $p5 = strpos($tmp2, 'language=');
                        if ($p5 !== false) {
                            $p5 += 10;
                            $char = substr($tmp2, ($p5 - 1), 1);
                            $p6 = strpos($tmp2, $char, $p5);
                            $lang = substr($tmp2, $p5, ($p6 - $p5));
                            # If language isn't "javascript", leave the tag alone
                            if (strtolower($lang) != 'javascript') {
                                continue;
                            }
                        }
                    }
                    # We now definitely have a javascript link tag, save it in an array
                    $url = '';
                    $p5 = strpos($tmp2, 'src=');
                    if ($p5 !== false) { # shouldn't happen, an address is needed
                        $p5 += 5;
                        $char = substr($tmp2, ($p5 - 1), 1);
                        $p6 = strpos($tmp2, $char, $p5);
                        $url = substr($tmp2, $p5, ($p6 - $p5));
                        # Clean the url from unnessecary path information (just leave the folder name like "js/xxx.js")
                        $url = str_replace($cfgClient[$client]['path']['htmlpath'], '', $url);
                        $url = (((substr($url, 0, 1) == '/') && (substr($url, 0, 2) != '//')) ? substr($url, 1) : $url);
                    }
                    # Exclude files with complete URLs or parameters from the compressor
                    $compress = (((substr($url, 0, 7) == 'http://') || (substr($url, 0, 8) == 'https://') || (substr($url, 0, 2) == '//') || (strpos($url, '?') !== false)) ? false : true);
                    $IEscripts[$c][] = array('url' => $url, 'compress' => $compress, 'old' => $tmp2);
                    $IEtmp[] = $tmp2;
                } else {
                    # Script block
                    $p4 = strpos($tmp, '</script>', $p3);
                    $tmp2 = substr($tmp, $p3, (($p4 - $p3) + 9));
                    $IEscriptblocks[$c][] = $tmp2;
                    $IEtmp[] = $tmp2;
                }
                $p3 = strpos($tmp, '<script', $p4);
            }
            
            # Find all style blocks
            $p3 = strpos($tmp, '<style');
            while ($p3 !== false) {
                $p4 = strpos($tmp, '</style>', $p3);
                $tmp2 = substr($tmp, $p3, (($p4 - $p3) + 8));
                $IEstyles[$c][] = $tmp2;
                $IEtmp[] = $tmp2;
                $p3 = strpos($tmp, '<style', $p4);
            }
            
            # Now that we found all links and blocks in this IE conditional comment, we delete them from it and see what's left
            $tmp2 = $tmp;
            for ($i = 0, $n = count($IEtmp); $i < $n; $i ++) {
                $tmp2 = str_replace($IEtmp[$i], '', $tmp2);
            }
            # Remove line breaks and tabs
            $tmp2 = str_replace(array("\r\n", "\r", "\n", "\t"), '', $tmp2);
            # Remove spaces
            while (strpos($tmp2, ' ') !== false) {
                $tmp2 = str_replace(' ', '', $tmp2);
            }
            # Now check the length of the resultung string
            if (strlen($tmp2) <= (strlen($c) + 16)) { # (start tag + end tag)
                # The block can be taken out from it's original position
                $IE[] = $tmp;
            } else {
                # We must take all found links and blocks out of this IE conditional comment
                for ($i = 0, $n = count($IEtmp); $i < $n; $i ++) {
                    $IE[] = $IEtmp[$i];
                }
            }
            # Move on to the next IE conditional comment block
            $p1 = strpos($sCode, '<!--[if', $p2);
        } else {
            $p1 = strpos($sCode, '<!--', ($p1 + 7));
        }
    }
    
    # Delete IE conditional comments and IE conditional comment entries marked as deletable
    for ($i = 0, $n = count($IE); $i < $n; $i ++) {
        $sCode = str_replace($IE[$i], '', $sCode);
    }
    
    # Find all link tags with rel="stylesheet", grouped by their media= parameter ({screen|all|print|...})
    $stylesheets = array();
    $iStyles = 0;
    $p1 = strpos($sCode, '<link');
    while ($p1 !== false) {
        $p2 = strpos($sCode, '>', $p1);
        $tmp = substr($sCode, $p1, (($p2 - $p1) + 1));
        if ((strpos(strtolower($tmp), 'rel="stylesheet"')) || (strpos(strtolower($tmp), "rel='stylesheet'"))) {
            # Extract the path and filename
            $url = '';
            $p3 = strpos($tmp, 'href=');
            if ($p3 !== false) { # shouldn't happen, an address is needed
                $p3 += 6;
                $char = substr($tmp, ($p3 - 1), 1);
                $p4 = strpos($tmp, $char, $p3);
                $url = substr($tmp, $p3, ($p4 - $p3));
                # Clean the url from unnessecary path information (just leave the folder name like "css/xxx.css")
                $url = str_replace($cfgClient[$client]['path']['htmlpath'], '', $url);
                $url = (((substr($url, 0, 1) == '/') && (substr($url, 0, 2) != '//')) ? substr($url, 1) : $url);
            }
            # Exclude files with complete URLs or parameters
            $compress = (((substr($url, 0, 7) == 'http://') || (substr($url, 0, 8) == 'https://') || (substr($url, 0, 2) == '//') || (strpos($url, '?') !== false)) ? false : true);
            if ($compress === false) {
                $iStyles ++;
            }
            # Sort into array grouped by media type
            $p3 = strpos($tmp, 'media=');
            if ($p3 !== false) {
                $p3 += 7;
                $char = substr($tmp, ($p3 - 1), 1);
                $p4 = strpos($tmp, $char, $p3);
                $media = substr($tmp, $p3, ($p4 - $p3));
                $stylesheets[$iStyles][$media][] = array('url' => $url, 'compress' => $compress, 'old' => $tmp);
            } else {
                $stylesheets[$iStyles]['all'][] = array('url' => $url, 'compress' => $compress, 'old' => $tmp);
            }
        }
        $p1 = strpos($sCode, '<link', $p2);
    }
    
    # Find all style blocks
    $styles = array();
    $p1 = strpos($sCode, '<style');
    while ($p1 !== false) {
        $p2 = strpos($sCode, '</style>', $p1);
        $tmp = substr($sCode, $p1, (($p2 - $p1) + 8));
        $styles[] = $tmp;
        $p1 = strpos($sCode, '<style', $p2);
    }
    
    if (!is_dir($sCacheDir . 'css')) {
        mkdir($sCacheDir . 'css', 0755, true);
    }
    
    # Rebuild the code for the non-IE-commented stylesheets
    foreach ($stylesheets as $iStyles => $aStylesheets) {
        foreach ($aStylesheets as $media => $sheets) {
            $files = array();
            $origfiles = array();
            $dtNewestFile = strtotime('1970-01-01 00:00:00');
            for ($i = 0, $n = count($sheets); $i < $n; $i ++) {
                $origfiles[] = basename($sheets[$i]['url']);
                $dtFileTime = filemtime($sheets[$i]['url']);
                if ($dtFileTime > $dtNewestFile) {
                    $dtNewestFile = $dtFileTime;
                }
                # Clean the url from unnessecary path information (just leave the folder name like "css/xxx.css")
                $sheets[$i]['url'] = str_replace($cfgClient[$client]['path']['htmlpath'], '', $sheets[$i]['url']);
                $sheets[$i]['url'] = (((substr($sheets[$i]['url'], 0, 1) == '/') && (substr($sheets[$i]['url'], 0, 2) != '//')) ? substr($sheets[$i]['url'], 1) : $sheets[$i]['url']);
                # Determine the filename, compiled if neccessary
                if ((substr($sheets[$i]['url'], -5) == '.less') && (is_file($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']))) {
                    if (!is_dir($sCacheDir . 'less_compiled')) {
                        mkdir($sCacheDir . 'less_compiled', 0755, true);
                    }
                    $sFileName = 'cache/less_compiled/' . Less_Cache::Get(array($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'] => dirname($sheets[$i]['url'])), array('cache_dir' => $sCacheDir . 'less_compiled/'));
                    rename($cfgClient[$client]['path']['frontend'] . $sFileName, $cfgClient[$client]['path']['frontend'] . $sFileName . '.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css');
                    $sFileName .= '.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css';
                    $c = file_get_contents($cfgClient[$client]['path']['frontend'] . $sFileName);
                    $c = str_replace(array("url('", "url('../../"), "url('../", $c);
                    $c = str_replace(array($cfgClient[$client]['path']['htmlpath'] . 'data:', $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:'), 'data:', $c);
                    file_put_contents($cfgClient[$client]['path']['frontend'] . $sFileName, $c);
                }
                elseif ((substr($sheets[$i]['url'], -5) == '.scss') && (is_file($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']))) {
                    if (!is_dir($sCacheDir . 'scss_compiled')) {
                        mkdir($sCacheDir . 'scss_compiled', 0755, true);
                    }
                    # Check if a compiled file with the same name exists in the cache dir, and if so, if it is newer than this file
                    $sFN = basename($sheets[$i]['url']);
                    $sFN = substr($sFN, 0, strrpos($sFN, '.')) . '.css.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css';
                    if ((file_exists($sCacheDir . 'scss_compiled/' . $sFN)) && (filemtime($sCacheDir . 'scss_compiled/' . $sFN) >= filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'])) && (filesize($sCacheDir . 'scss_compiled/' . $sFN) > 0)) {
                        $sFileName = 'cache/scss_compiled/' . $sFN;
                    }
                    else {
                        cInclude('external', 'Scss/Base/Range.php');
                        cInclude('external', 'Scss/Block.php');
                        cInclude('external', 'Scss/Colors.php');
                        cInclude('external', 'Scss/Compiler.php');
                        cInclude('external', 'Scss/Compiler/Environment.php');
                        cInclude('external', 'Scss/Exception/CompilerException.php');
                        cInclude('external', 'Scss/Exception/ParserException.php');
                        cInclude('external', 'Scss/Exception/RangeException.php');
                        cInclude('external', 'Scss/Exception/ServerException.php');
                        cInclude('external', 'Scss/Formatter.php');
                        cInclude('external', 'Scss/Formatter/Compact.php');
                        cInclude('external', 'Scss/Formatter/Compressed.php');
                        cInclude('external', 'Scss/Formatter/Crunched.php');
                        cInclude('external', 'Scss/Formatter/Debug.php');
                        cInclude('external', 'Scss/Formatter/Expanded.php');
                        cInclude('external', 'Scss/Formatter/Nested.php');
                        cInclude('external', 'Scss/Formatter/OutputBlock.php');
                        cInclude('external', 'Scss/Node.php');
                        cInclude('external', 'Scss/Node/Number.php');
                        cInclude('external', 'Scss/Parser.php');
                        cInclude('external', 'Scss/SourceMap/Base64.php');
                        cInclude('external', 'Scss/SourceMap/Base64VLQ.php');
                        cInclude('external', 'Scss/SourceMap/SourceMapGenerator.php');
                        cInclude('external', 'Scss/Type.php');
                        cInclude('external', 'Scss/Util.php');
                        cInclude('external', 'Scss/Version.php');
                        try {
                            $scss = new Leafo\ScssPhp\Compiler();
                            $scss->setFormatter('Leafo\ScssPhp\Formatter\Expanded');
                            $scss->setImportPaths($cfgClient[$client]['path']['frontend'] . dirname($sheets[$i]['url']) . '/');
                            $sCompiled = $scss->compile(file_get_contents($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']));
                            $sCompiled = str_replace(array($cfgClient[$client]['path']['htmlpath'] . 'data:', $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'), 'data:', $sCompiled);
                            if (file_put_contents($cfgClient[$client]['path']['frontend'] . 'cache/scss_compiled/' . $sFN, $sCompiled) !== false) {
                                $sFileName = 'cache/scss_compiled/' . $sFN;
                            }
                            else {
                                $sFN = 'cache/css/' . basename($sheets[$i]['url']) . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . 'css';
                                copy($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'], $cfgClient[$client]['path']['frontend'] . $sFN);
                                $sFileName = $sFN;
                            }
                            unset($scss);
                        } catch (\Exception $e) {
                            $refl = new ReflectionClass($e);
                            $prop = $refl->getProperty('message');
                            $prop->setAccessible(true);
                            $p1 = strpos($sCode, '<body');
                            $p2 = (strpos($sCode, '>', $p1) + 1);
                            $sCode = substr($sCode, 0, $p2) . "\n" . '<h6>SCSS error: ' . $prop->getValue($e) . '</h6>' . substr($sCode, $p2);
                            $sFN = 'cache/css/' . basename($sheets[$i]['url']) . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . 'css';
                            copy($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'], $cfgClient[$client]['path']['frontend'] . $sFN);
                            touch($cfgClient[$client]['path']['frontend'] . $sFN, filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']));
                            $sFileName = $sFN;
                        }
                    }
                }
                else {
                    $sFN = 'cache/css/' . basename($sheets[$i]['url']) . '.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css';
                    if (copy($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'], $cfgClient[$client]['path']['frontend'] . $sFN)) {
                        touch($cfgClient[$client]['path']['frontend'] . $sFN, filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']));
                        $sFileName = $sFN;
                    }
                    else {
                        $sFileName = $sheets[$i]['url'];
                    }
                }
                # Correct the paths of linked files ('url("...")')
                $fmtime = filemtime($cfgClient[$client]['path']['frontend'] . $sFileName);
                $cnt = file_get_contents($cfgClient[$client]['path']['frontend'] . $sFileName);
                $cnt = _correctLinkedPaths($cnt, dirname(realpath($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'])), 2);
                if (strpos($switch, 'images') !== false) {
                    $cnt = _convertImagesFromCodeToWebP($cnt, dirname(realpath($cfgClient[$client]['path']['frontend'] . $sFileName)), 2);
                }
                $cnt = str_replace(
                    array(
                        $cfgClient[$client]['path']['htmlpath'] . 'data:',
                        $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:',
                        $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'
                    ),
                    'data:',
                    $cnt
                );
                file_put_contents($cfgClient[$client]['path']['frontend'] . $sFileName, $cnt);
                touch($cfgClient[$client]['path']['frontend'] . $sFileName, $fmtime);
                # Correct or compress?
                if ((strpos($switch, 'correct') !== false) || ($sheets[$i]['compress'] == false)) {
                    # Remove the old links for this media type
                    $sCode = str_replace($sheets[$i]['old'], '', $sCode);
                    # Copy the file to the client's cache directory
                    copy($cfgClient[$client]['path']['frontend'] . $sFileName, $cfgClient[$client]['path']['frontend'] . 'cache/' . basename($sFileName));
                    # Correct the linked file paths
                    $fmtime = filemtime($cfgClient[$client]['path']['frontend'] . 'cache/' . basename($sFileName));
                    $cnt = file_get_contents($cfgClient[$client]['path']['frontend'] . 'cache/' . basename($sFileName));
                    $cnt = _correctLinkedPaths($cnt, dirname(realpath($cfgClient[$client]['path']['frontend'] . $sFileName)), 1);
                    $sFileName = 'cache/' . basename($sFileName);
                    $cnt = str_replace(
                        array(
                            $cfgClient[$client]['path']['htmlpath'] . 'data:',
                            $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:',
                            $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'
                        ),
                        'data:',
                        $cnt
                    );
                    file_put_contents($cfgClient[$client]['path']['frontend'] . $sFileName, $cnt);
                    touch($cfgClient[$client]['path']['frontend'] . $sFileName, $fmtime);
                    # Add the links to the end of the head section
                    $p1 = strpos($sCode, '</head>');
                    $sCode = substr($sCode, 0, $p1) . '<link rel="stylesheet" type="text/css" media="' . $media . '" href="' . $sFileName . '" />' . "\n" . substr($sCode, $p1);
                }
                else {
                    # Call the compressor class for each media type and replace the old links with the single new one
                    # Find and extract included stylesheet files
                    if (is_file($cfgClient[$client]['path']['frontend'] . $sFileName)) {
                        $fmtime = filemtime($cfgClient[$client]['path']['frontend'] . $sFileName);
                        $content = @file_get_contents($cfgClient[$client]['path']['frontend'] . $sFileName);
                        if (strrpos($sFileName, '/') > 0) {
                            $path = substr($sFileName, 0, (strrpos($sFileName, '/') + 1));
                        } else {
                            $path = '';
                        }
                        $p1 = strpos(strtolower($content), '@import url(');
                        while ($p1 !== false) {
                            $p2 = strpos($content, ');', $p1);
                            $url = str_replace(array('"', "'"), '', substr($content, ($p1 + 12), ($p2 - $p1 - 12)));
                            # Clean the url from unnessecary path information (just leave the folder name like "css/xxx.css")
                            $url = str_replace($cfgClient[$client]['path']['htmlpath'], '', $url);
                            $url = (((substr($url, 0, 1) == '/') && (substr($url, 0, 2) != '//')) ? substr($url, 1) : $url);
                            # Exclude files with complete URLs or parameters
                            if ((substr($url, 0, 7) != 'http://') && (substr($url, 0, 8) != 'https://') && (substr($url, 0, 2) != '//') && (strpos($url, '?') === false)) {
                                $files[] = $cfgClient[$client]['path']['frontend'] . $path . $url;
                                $origfiles[] = basename($url);
                            } else {
                                # Files with absolute paths must be placed in the HTML head section
                                $p1 = strpos($sCode, '</head>');
                                $sCode = substr($sCode, 0, $p1) . '<link rel="stylesheet" type="text/css" media="' . $media . '" href="' . $path . $url . '" />' . "\n" . substr($sCode, $p1);
                            }
                            $content = substr($content, 0, $p1) . substr($content, ($p2 + 2));
                            $p1 = strpos(strtolower($content), '@import url(');
                        }
                        $content = str_replace(
                            array(
                                $cfgClient[$client]['path']['htmlpath'] . 'data:',
                                $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:',
                                $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'
                            ),
                            'data:',
                            $content
                        );
                        file_put_contents($cfgClient[$client]['path']['frontend'] . $sFileName, $content);
                        touch($cfgClient[$client]['path']['frontend'] . $sFileName, $fmtime);
                        # Is there any content left in this file?
                        if (strlen(trim(str_replace(array("\r", "\n"), '', $content)))) {
                            # Yes, add it to the compressor, which will include it's remaining content in the compressed file
                            $files[] = $cfgClient[$client]['path']['frontend'] . $sFileName;
                        }
                    }
                }
            }
            if (count($files)) {
                # Get a name for the compressed file
                $md5 = md5(implode('', $origfiles));
                # Compress the files into a single one
                $compressed = Output_Compressor::generate($cfgClient[$client]['path']['frontend'] . 'cache/', $files, 'css', $cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $md5, $dtNewestFile);
                if ($compressed) {
                    # Correct the linked file paths
                    $fmtime = filemtime($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed);
                    $cnt = file_get_contents($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed);
                    $cnt = _correctLinkedPaths($cnt, dirname($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed), 1);
                    $cnt = str_replace(
                        array(
                            $cfgClient[$client]['path']['htmlpath'] . 'data:',
                            $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:',
                            $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'
                        ),
                        'data:',
                        $cnt
                    );
                    file_put_contents($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed, $cnt);
                    touch($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed, $fmtime);
                    # Add the compressed file link to the end of the head section
                    $p1 = strpos($sCode, '</head>');
                    $sCode = substr($sCode, 0, $p1) . '<link rel="stylesheet" type="text/css" media="' . $media . '" href="front_content.php?action=get_compressed&f=' . $compressed . '&c=css" />' . "\n" . substr($sCode, $p1);
                }
                foreach ($files as $file) {
                    if (substr($file, 0, strlen($cfgClient[$client]['path']['frontend'] . 'cache/css/')) == $cfgClient[$client]['path']['frontend'] . 'cache/css/') {
                        unlink($file);
                    }
                    if (substr($file, 0, strlen($cfgClient[$client]['path']['frontend'] . 'cache/scss_compiled/')) == $cfgClient[$client]['path']['frontend'] . 'cache/scss_compiled/') {
                        unlink($file);
                    }
                }
            }
            # Remove the old links for this media type
            for ($i = 0, $n = count($sheets); $i < $n; $i ++) {
                if (strlen($sFileName)) {
                    $sCode = str_replace($sheets[$i]['old'], '', $sCode);
                }
            }
        }
    }
    unset($stylesheets);
    unset($aStylesheets);
    unset($iStyles);
    unset($media);
    unset($sheets);
    
    # Rebuild the code for the IE-commented stylesheets
    if (count($IElinks)) {
        foreach ($IElinks as $condition => $links) {
            # Add the conditional comment start tag to the end of the head section
            $p1 = strpos($sCode, '</head>');
            $sCode = substr($sCode, 0, $p1) . $condition . "\n" . substr($sCode, $p1);
            # Put the links there
            foreach ($links as $media => $sheets) {
                $files = array();
                $origfiles = array();
                $dtNewestFile = strtotime('1970-01-01 00:00:00');
                for ($i = 0, $n = count($sheets); $i < $n; $i ++) {
                    $origfiles[] = basename($sheets[$i]['url']);
                    $dtFileTime = filemtime($sheets[$i]['url']);
                    if ($dtFileTime > $dtNewestFile) {
                        $dtNewestFile = $dtFileTime;
                    }
                    # Clean the url from unnessecary path information (just leave the folder name like "css/xxx.css")
                    $sheets[$i]['url'] = str_replace($cfgClient[$client]['path']['htmlpath'], '', $sheets[$i]['url']);
                    $sheets[$i]['url'] = (((substr($sheets[$i]['url'], 0, 1) == '/') && (substr($sheets[$i]['url'], 0, 2) != '//')) ? substr($sheets[$i]['url'], 1) : $sheets[$i]['url']);
                    # Determine the filename, compiled if neccessary
                    if ((substr($sheets[$i]['url'], -5) == '.less') && (is_file($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']))) {
                        if (!is_dir($sCacheDir . 'less_compiled')) {
                            mkdir($sCacheDir . 'less_compiled', 0755, true);
                        }
                        $sFileName = 'cache/less_compiled/' . Less_Cache::Get(array($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'] => dirname($sheets[$i]['url'])), array('cache_dir' => $sCacheDir . 'less_compiled/'));
                        rename($cfgClient[$client]['path']['frontend'] . $sFileName, $cfgClient[$client]['path']['frontend'] . $sFileName . '.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css');
                        $sFileName .= '.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css';
                            $c = file_get_contents($cfgClient[$client]['path']['frontend'] . $sFileName);
                        $c = str_replace(array("url('", "url('../../"), "url('../", $c);
                        $c = str_replace(array($cfgClient[$client]['path']['htmlpath'] . 'data:', $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:'), 'data:', $c);
                        file_put_contents($cfgClient[$client]['path']['frontend'] . $sFileName, $c);
                    }
                    elseif ((substr($sheets[$i]['url'], -5) == '.scss') && (is_file($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']))) {
                        if (!is_dir($sCacheDir . 'scss_compiled')) {
                            mkdir($sCacheDir . 'scss_compiled', 0755, true);
                        }
                        # Check if a compiled file with the same name exists in the cache dir, and if so, if it is newer than this file
                        $sFN = basename($sheets[$i]['url']);
                        $sFN = substr($sFN, 0, strrpos($sFN, '.')) . '.css.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css';
                        if ((file_exists($sCacheDir . 'scss_compiled/' . $sFN)) && (filemtime($sCacheDir . 'scss_compiled/' . $sFN) >= filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'])) && (filesize($sCacheDir . 'scss_compiled/' . $sFN) > 0)) {
                            $sFileName = 'cache/scss_compiled/' . $sFN;
                        }
                        else {
                            cInclude('external', 'Scss/Base/Range.php');
                            cInclude('external', 'Scss/Block.php');
                            cInclude('external', 'Scss/Colors.php');
                            cInclude('external', 'Scss/Compiler.php');
                            cInclude('external', 'Scss/Compiler/Environment.php');
                            cInclude('external', 'Scss/Exception/CompilerException.php');
                            cInclude('external', 'Scss/Exception/ParserException.php');
                            cInclude('external', 'Scss/Exception/RangeException.php');
                            cInclude('external', 'Scss/Exception/ServerException.php');
                            cInclude('external', 'Scss/Formatter.php');
                            cInclude('external', 'Scss/Formatter/Compact.php');
                            cInclude('external', 'Scss/Formatter/Compressed.php');
                            cInclude('external', 'Scss/Formatter/Crunched.php');
                            cInclude('external', 'Scss/Formatter/Debug.php');
                            cInclude('external', 'Scss/Formatter/Expanded.php');
                            cInclude('external', 'Scss/Formatter/Nested.php');
                            cInclude('external', 'Scss/Formatter/OutputBlock.php');
                            cInclude('external', 'Scss/Node.php');
                            cInclude('external', 'Scss/Node/Number.php');
                            cInclude('external', 'Scss/Parser.php');
                            cInclude('external', 'Scss/SourceMap/Base64.php');
                            cInclude('external', 'Scss/SourceMap/Base64VLQ.php');
                            cInclude('external', 'Scss/SourceMap/SourceMapGenerator.php');
                            cInclude('external', 'Scss/Type.php');
                            cInclude('external', 'Scss/Util.php');
                            cInclude('external', 'Scss/Version.php');
                            try {
                                $scss = new Leafo\ScssPhp\Compiler();
                                $scss->setFormatter('Leafo\ScssPhp\Formatter\Expanded');
                                $scss->setImportPaths($cfgClient[$client]['path']['frontend'] . dirname($sheets[$i]['url']) . '/');
                                $sCompiled = $scss->compile(file_get_contents($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']));
                                $sCompiled = str_replace(array($cfgClient[$client]['path']['htmlpath'] . 'data:', $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'), 'data:', $sCompiled);
                                if (file_put_contents($cfgClient[$client]['path']['frontend'] . 'cache/scss_compiled/' . $sFN, $sCompiled) !== false) {
                                    $sFileName = 'cache/scss_compiled/' . $sFN;
                                }
                                else {
                                    $sFileName = $sheets[$i]['url'];
                                }
                                unset($scss);
                            } catch (\Exception $e) {
                                $refl = new ReflectionClass($e);
                                $prop = $refl->getProperty('message');
                                $prop->setAccessible(true);
                                $p1 = strpos($sCode, '<body');
                                $p2 = (strpos($sCode, '>', $p1) + 1);
                                $sCode = substr($sCode, 0, $p2) . "\n" . '<h6>SCSS error: ' . $prop->getValue($e) . '</h6>' . substr($sCode, $p2);
                                $sFN = 'cache/css/' . basename($sheets[$i]['url']) . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . 'css';
                                copy($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'], $cfgClient[$client]['path']['frontend'] . $sFN);
                                touch($cfgClient[$client]['path']['frontend'] . $sFN, filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']));
                                $sFileName = $sFN;
                            }
                        }
                    }
                    else {
                        $sFN = 'cache/css/' . basename($sheets[$i]['url']) . '.' . filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']) . '.css';
                        if (copy($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url'], $cfgClient[$client]['path']['frontend'] . $sFN)) {
                            touch($cfgClient[$client]['path']['frontend'] . $sFN, filemtime($cfgClient[$client]['path']['frontend'] . $sheets[$i]['url']));
                            $sFileName = $sFN;
                        }
                        else {
                            $sFileName = $sheets[$i]['url'];
                        }
                    }
                        # Correct or compress?
                    if (strpos($switch, 'correct') !== false) {
                        # Remove the old links for this media type
                        if (strlen($sFileName)) {
                            $sCode = str_replace($sheets[$i]['old'], '', $sCode);
                        }
                        # Add the links to the end of the head section
                        $p1 = strpos($sCode, '</head>');
                        if (strlen($sFileName)) {
                            $sCode = substr($sCode, 0, $p1) . '<link rel="stylesheet" type="text/css" media="' . $media . '" href="' . $sFileName . '" />' . "\n" . substr($sCode, $p1);
                        }
                    }
                    else {
                        # Call the compressor class for each media type and replace the old links with the single new one
                        if (strlen($sFileName)) {
                            # Find and extract included stylesheet files
                            if (is_file($cfgClient[$client]['path']['frontend'] . $sFileName)) {
                                $content = @file_get_contents($cfgClient[$client]['path']['frontend'] . $sFileName);
                                if (strrpos($sFileName, '/') > 0) {
                                    $path = substr($sFileName, 0, (strrpos($sFileName, '/') + 1));
                                } else {
                                    $path = '';
                                }
                                $p1 = strpos(strtolower($content), '@import url(');
                                while ($p1 !== false) {
                                    $p2 = strpos($content, ');', $p1);
                                    $url = str_replace(array('"', "'"), '', substr($content, ($p1 + 12), ($p2 - $p1 - 12)));
                                    # Clean the url from unnessecary path information (just leave the folder name like "css/xxx.css")
                                    $url = str_replace($cfgClient[$client]['path']['htmlpath'], '', $url);
                                    $url = (((substr($url, 0, 1) == '/') && (substr($url, 0, 2) != '//')) ? substr($url, 1) : $url);
                                    # Exclude files with complete URLs or parameters
                                    if ((substr($url, 0, 7) != 'http://') && (substr($url, 0, 8) != 'https://') && (substr($url, 0, 2) != '//') && (strpos($url, '?') === false)) {
                                        $files[] = $path . $url;
                                        $origfiles[] = basename($url);
                                    } else {
                                        # Files with absolute paths must be placed in the HTML head section
                                        $p1 = strpos($sCode, '</head>');
                                        $sCode = substr($sCode, 0, $p1) . '<link rel="stylesheet" type="text/css" media="' . $media . '" href="' . $path . $url . '" />' . "\n" . substr($sCode, $p1);
                                    }
                                    $content = substr($content, 0, $p1) . substr($content, ($p2 + 2));
                                    $p1 = strpos(strtolower($content), '@import url(');
                                }
                                # Is there any content left in this file?
                                if (strlen(trim(str_replace(array("\r", "\n"), '', $content)))) {
                                    # Yes, add it to the compressor, which will include it's remaining content in the compressed file
                                    $files[] = $sFileName;
                                }
                            }
                        }
                    }
                }
                if (count($files)) {
                    # Get a name for the compressed file
                    $md5 = md5(implode('', $origfiles));
                    # Make the paths absolute for the compressor
                    for ($i = 0, $n = count($files); $i < $n; $i ++) {
                        $files[$i] = $cfgClient[$client]['path']['frontend'] . $files[$i];
                    }
                    # Compress the files into a single one
                    $compressed = Output_Compressor::generate($cfgClient[$client]['path']['frontend'] . 'cache/', $files, 'css', $cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $md5, $dtNewestFile);
                    if ($compressed) {
                        # Correct the linked file paths
                        $fmtime = filemtime($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed);
                        $cnt = file_get_contents($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed);
                        $cnt = _correctLinkedPaths($cnt, dirname($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed), 1);
                        $cnt = str_replace(
                            array(
                                $cfgClient[$client]['path']['htmlpath'] . 'data:',
                                $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:',
                                $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'
                            ),
                            'data:',
                            $cnt
                        );
                        file_put_contents($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed, $cnt);
                        touch($cfgClient[$client]['path']['frontend'] . 'cache/' . $compressed, $fmtime);
                        # Add the compressed file link to the end of the head section
                        $p1 = strpos($sCode, '</head>');
                        $sCode = substr($sCode, 0, $p1) . '<link rel="stylesheet" type="text/css" media="' . $media . '" href="front_content.php?action=get_compressed&f=' . $compressed . '&c=css" />' . "\n" . substr($sCode, $p1);
                    }
                    foreach ($files as $file) {
                        if (substr($file, 0, strlen($cfgClient[$client]['path']['frontend'] . 'cache/')) == $cfgClient[$client]['path']['frontend'] . 'cache/') {
                            unlink($file);
                        }
                    }
                }
                # Remove the old links for this media type
                for ($i = 0, $n = count($sheets); $i < $n; $i ++) {
                    if (strlen($sFileName)) {
                        $sCode = str_replace($sheets[$i]['old'], '', $sCode);
                    }
                }
            }
            # Add the conditional comment end tag to the end of the head section
            $p1 = strpos($sCode, '</head>');
            $sCode = substr($sCode, 0, $p1) . ((substr($condition, -5) == '<!-->') ? '<!--' : '') . '<![endif]-->' . "\n" . substr($sCode, $p1);
        }
    }
    unset($IElinks);
    unset($condition);
    unset($links);
    unset($media);
    unset($sheets);
    unset($files);
    
    # Rebuild the code for the style blocks
    if (count($styles)) {
        for ($i = 0, $n = count($styles); $i < $n; $i ++) {
            $sCode = str_replace($styles[$i], '', $sCode);
            $p1 = strpos($sCode, '</head>');
            $sCode = substr($sCode, 0, $p1) . $styles[$i] . "\n" . substr($sCode, $p1);
        }
    }
    unset($styles);
    
    # Rebuild the code for the IE-commented style blocks
    if (count($IEstyles)) {
        foreach ($IEstyles as $condition => $styles) {
            # Add the conditional comment start tag to the end of the head section
            $p1 = strpos($sCode, '</head>');
            $sCode = substr($sCode, 0, $p1) . $condition . "\n" . substr($sCode, $p1);
            # Put the style blocks there
            for ($i = 0, $n = count($styles); $i < $n; $i ++) {
                $p1 = strpos($sCode, '</head>');
                $sCode = substr($sCode, 0, $p1) . $styles[$i] . "\n" . substr($sCode, $p1);
            }
            # Add the conditional comment end tag to the end of the head section
            $p1 = strpos($sCode, '</head>');
            $sCode = substr($sCode, 0, $p1) . '<![endif]-->' . "\n" . substr($sCode, $p1);
        }
    }
    unset($IEstyles);
    unset($condition);
    unset($styles);
    
    # Find all script link tags and script blocks
    $scripts = array();
    $scriptblocks = array();
    $p1 = strpos($sCode, '<script');
    while ($p1 !== false) {
        $p2 = strpos($sCode, '>', $p1);
        # Check if this is included code, or code directly in the page
        if (substr($sCode, $p2, 10) == '></script>') {
            # Script link
            $tmp = substr($sCode, $p1, (($p2 - $p1) + 10));
            # Check if this script tag has a type or language set
            $p3 = strpos($tmp, 'type=');
            if ($p3 !== false) {
                $p3 += 6;
                $char = substr($tmp, ($p3 - 1), 1);
                $p4 = strpos($tmp, $char, $p3);
                $type = substr($tmp, $p3, ($p4 - $p3));
                # If type isn't "text/javascript", leave the tag alone
                if (strtolower($type) != 'text/javascript') {
                    $p1 = strpos($sCode, '<script', $p2);
                    continue;
                }
            } else {
                $p3 = strpos($tmp, 'language=');
                if ($p3 !== false) {
                    $p3 += 10;
                    $char = substr($tmp, ($p3 - 1), 1);
                    $p4 = strpos($tmp, $char, $p3);
                    $lang = substr($tmp, $p3, ($p4 - $p3));
                    # If language isn't "javascript", leave the tag alone
                    if (strtolower($lang) != 'javascript') {
                        $p1 = strpos($sCode, '<script', $p2);
                        continue;
                    }
                }
            }
            # We now definitely have a javascript link tag, save it in an array
            $url = '';
            $p3 = strpos($tmp, 'src=');
            if ($p3 !== false) { # shouldn't happen, an address is needed
                $p3 += 5;
                $char = substr($tmp, ($p3 - 1), 1);
                $p4 = strpos($tmp, $char, $p3);
                $url = substr($tmp, $p3, ($p4 - $p3));
                # Clean the url from unnessecary path information (just leave the folder name like "js/xxx.js")
                $url = str_replace($cfgClient[$client]['path']['htmlpath'], '', $url);
                $url = (((substr($url, 0, 1) == '/') && (substr($url, 0, 2) != '//')) ? substr($url, 1) : $url);
            }
            # Exclude files with complete URLs or parameters from the compressor
            $compress = (((substr($url, 0, 7) == 'http://') || (substr($url, 0, 8) == 'https://') || (substr($url, 0, 2) == '//') || (strpos($url, '?') !== false)) ? false : true);
            $scripts[] = array('url' => $url, 'compress' => $compress, 'old' => $tmp);
        } else {
            # Script block
            $p2 = strpos($sCode, '</script>', $p1);
            $tmp = substr($sCode, $p1, (($p2 - $p1) + 9));
            $scriptblocks[] = $tmp;
            #$p1 = strpos($sCode, '<script', ($p1 + 1));
        }
        $p1 = strpos($sCode, '<script', ($p1 + 1));
    }
    
    # Delete the old script links from the page code
    if (count($scripts)) {
        for ($i = 0, $n = count($scripts); $i < $n; $i ++) {
            # Remove the old link
            $sCode = str_replace($scripts[$i]['old'], '', $sCode);
        }
    }
    
    # Delete the old script blocks from the page code
    if (count($scriptblocks)) {
        for ($i = 0, $n = count($scriptblocks); $i < $n; $i ++) {
            # Remove the old script block
            $sCode = str_replace($scriptblocks[$i], '', $sCode);
        }
    }
    
    # Call the compressor class and replace the old script links with the single new one
    $files = array();
    if (count($scripts)) {
        for ($i = 0, $n = count($scripts); $i < $n; $i ++) {
            # Correct or compress?
            if ((strpos($switch, 'correct') !== false) || ($scripts[$i]['compress'] == false)) {
                # First, if we have files in our array, compress them and output them as a single file (needed to presave the loading order)
                if (count($files)) {
                    $dtNewestFile = strtotime('1970-01-01 00:00:00');
                    for ($i = 0, $n = count($files); $i < $n; $i ++) {
                        $dtFileTime = filemtime($files[$i]);
                        if ($dtFileTime > $dtNewestFile) {
                            $dtNewestFile = $dtFileTime;
                        }
                    }
                    # Get a name for the compressed file
                    $md5 = md5(implode('', $files));
                    # Delete previously generated files with the md5 name
                    array_map('unlink', glob($cfgClient[$client]['path']['frontend'] . 'cache/' . substr($md5, 0, -14) . '*.js*'));
                    # Compress the files into one single file
                    $compressed = Output_Compressor::generate($cfgClient[$client]['path']['frontend'] . 'cache/', $files, 'js', $cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $md5, $dtNewestFile);
                    if ($compressed) {
                        $tmp = '<script type="text/javascript" src="front_content.php?action=get_compressed&f=' . $compressed . '&c=javascript"></script>';
                        # Add the link to the end of the body section
                        $p1 = strpos($sCode, '</body>');
                        $sCode = substr($sCode, 0, $p1) . $tmp . "\n" . substr($sCode, $p1);
                        $files = array();
                    }
                }
                # Create the link code
                $tmp = '<script type="text/javascript" src="' . $scripts[$i]['url'] . '"></script>';
                # Add the link to the end of the body section
                $p1 = strpos($sCode, '</body>');
                $sCode = substr($sCode, 0, $p1) . $tmp . "\n" . substr($sCode, $p1);
            } else {
                $files[] = $cfgClient[$client]['path']['frontend'] . $scripts[$i]['url'];
            }
        }
        if (count($files)) {
            $dtNewestFile = strtotime('1970-01-01 00:00:00');
            for ($i = 0, $n = count($files); $i < $n; $i ++) {
                $dtFileTime = filemtime($files[$i]);
                if ($dtFileTime > $dtNewestFile) {
                    $dtNewestFile = $dtFileTime;
                }
            }
            # Get a name for the compressed file
            $md5 = md5(implode('', $files));
            # Delete previously generated files with the md5 name
            array_map('unlink', glob($cfgClient[$client]['path']['frontend'] . 'cache/' . substr($md5, 0, -14) . '*.js*'));
            # Compress the files into one single file
            $compressed = Output_Compressor::generate($cfgClient[$client]['path']['frontend'] . 'cache/', $files, 'js', $cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $md5, $dtNewestFile);
            if ($compressed) {
                $tmp = '<script type="text/javascript" src="front_content.php?action=get_compressed&f=' . $compressed . '&c=javascript"></script>';
                # Add the link to the end of the body section
                $p1 = strpos($sCode, '</body>');
                $sCode = substr($sCode, 0, $p1) . $tmp . "\n" . substr($sCode, $p1);
                $files = array();
            }
        }
    }
    unset($scripts);
    unset($type);
    unset($lang);
    unset($files);
    
    # Rebuild the code for the script blocks
    if (count($scriptblocks)) {
        for ($i = 0, $n = count($scriptblocks); $i < $n; $i ++) {
            $sCode = str_replace($scriptblocks[$i], '', $sCode);
            $p1 = strpos($sCode, '</body>');
            $sCode = substr($sCode, 0, $p1) . $scriptblocks[$i] . "\n" . substr($sCode, $p1);
        }
    }
    unset($scriptblocks);
    
    # Rebuild the code for the IE-commented script links
    if (count($IEscripts)) {
        foreach ($IEscripts as $condition => $links) {
            # Add the conditional comment start tag to the end of the body section
            $p1 = strpos($sCode, '</body>');
            $sCode = substr($sCode, 0, $p1) . $condition . "\n" . substr($sCode, $p1);
            # Put the links there
            # Call the compressor class and replace the old script links with the single new one
            $files = array();
            if (count($links)) {
                for ($i = 0, $n = count($links); $i < $n; $i ++) {
                    # Correct or compress?
                    if ((strpos($switch, 'correct') !== false) || ($links[$i]['compress'] == false)) {
                        # First, if we have files in our array, compress them and output them as a single file (needed to presave the loading order)
                        if (count($files)) {
                            $dtNewestFile = strtotime('1970-01-01 00:00:00');
                            for ($i = 0, $n = count($files); $i < $n; $i ++) {
                                $dtFileTime = filemtime($files[$i]);
                                if ($dtFileTime > $dtNewestFile) {
                                    $dtNewestFile = $dtFileTime;
                                }
                            }
                            # Get a name for the compressed file
                            $md5 = md5(implode('', $files));
                            # Delete previously generated files with the md5 name
                            array_map('unlink', glob($cfgClient[$client]['path']['frontend'] . 'cache/' . substr($md5, 0, -14) . '*.js*'));
                            # Compress the files into one single file
                            $compressed = Output_Compressor::generate($cfgClient[$client]['path']['frontend'] . 'cache/', $files, 'js', $cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $md5, $dtNewestFile);
                            if ($compressed) {
                                $tmp = '<script type="text/javascript" src="front_content.php?action=get_compressed&f=' . $compressed . '&c=javascript"></script>';
                                # Add the link to the end of the body section
                                $p1 = strpos($sCode, '</body>');
                                $sCode = substr($sCode, 0, $p1) . $tmp . "\n" . substr($sCode, $p1);
                                $files = array();
                            }
                        }
                        # Create the link code
                        $tmp = '<script type="text/javascript" src="' . $links[$i]['url'] . '"></script>';
                        # Add the link to the end of the body section
                        $p1 = strpos($sCode, '</body>');
                        $sCode = substr($sCode, 0, $p1) . $tmp . "\n" . substr($sCode, $p1);
                    } else {
                        $files[] = $cfgClient[$client]['path']['frontend'] . $links[$i]['url'];
                    }
                }
                if (count($files)) {
                    $dtNewestFile = strtotime('1970-01-01 00:00:00');
                    for ($i = 0, $n = count($files); $i < $n; $i ++) {
                        $dtFileTime = filemtime($files[$i]);
                        if ($dtFileTime > $dtNewestFile) {
                            $dtNewestFile = $dtFileTime;
                        }
                    }
                    # Get a name for the compressed file
                    $md5 = md5(implode('', $files));
                    # Delete previously generated files with the md5 name
                    array_map('unlink', glob($cfgClient[$client]['path']['frontend'] . 'cache/' . substr($md5, 0, -14) . '*.js*'));
                    # Compress the files into one single file
                    $compressed = Output_Compressor::generate($cfgClient[$client]['path']['frontend'] . 'cache/', $files, 'js', $cfgClient[$client]['path']['htmlpath'], $cfgClient[$client]['path']['frontend'], $md5, $dtNewestFile);
                    if ($compressed) {
                        $tmp = '<script type="text/javascript" src="front_content.php?action=get_compressed&f=' . $compressed . '&c=javascript"></script>';
                        # Add the link to the end of the body section
                        $p1 = strpos($sCode, '</body>');
                        $sCode = substr($sCode, 0, $p1) . $tmp . "\n" . substr($sCode, $p1);
                        $files = array();
                    }
                }
            }
            # Add the conditional comment end tag to the end of the body section
            $p1 = strpos($sCode, '</body>');
            $sCode = substr($sCode, 0, $p1) . '<![endif]-->' . "\n" . substr($sCode, $p1);
        }
    }
    unset($IEscripts);
    unset($condition);
    unset($links);
    
    # Rebuild the code for the IE-commented script blocks
    if (count($IEscriptblocks)) {
        foreach ($IEscriptblocks as $condition => $link) {
            if (is_array($link)) {
                $p1 = strpos($sCode, '</body>');
                $sCode = substr($sCode, 0, $p1) . $condition . "\n" . substr($sCode, $p1);
                foreach ($link as $lnk) {
                    $p1 = strpos($sCode, '</body>');
                    $sCode = substr($sCode, 0, $p1) . $lnk . "\n" . substr($sCode, $p1);
                }
                $p1 = strpos($sCode, '</body>');
                $sCode = substr($sCode, 0, $p1) . ((substr($condition, -5) == '<!-->') ? '<!--' : '') . '<![endif]-->' . "\n" . substr($sCode, $p1);
            } else {
                # Add the conditional comment tag to the end of the body section
                $p1 = strpos($sCode, '</body>');
                $sCode = substr($sCode, 0, $p1) . $condition . "\n" . $link . "\n" . '<![endif]-->' . "\n" . substr($sCode, $p1);
            }
        }
    }
    unset($IEscriptblocks);
    unset($condition);
    unset($link);
    
    # Find all link tags with rel="preload"
    $preloads = array();
    $iStyles = 0;
    $p1 = strpos($sCode, '<link');
    while ($p1 !== false) {
        $p2 = strpos($sCode, '>', $p1);
        $tmp = substr($sCode, $p1, (($p2 - $p1) + 1));
        if ((strpos(strtolower($tmp), 'rel="preload"')) || (strpos(strtolower($tmp), "rel='preload'"))) {
#$sDbgCode .= '<!-- ' . $tmp . ' -->' . "\n";
            # Remember the link
            $preloads[] = $tmp;
            # Delete the link in the html code
            $sCode = str_replace($tmp, '', $sCode);
            $p1 = strpos($sCode, '<link', $p1);
        }
        else {
            $p1 = strpos($sCode, '<link', $p2);
        }
    }
    
    # Insert the Reload link tags at the end of the <head> section
    foreach ($preloads as $preload) {
        $sCode = str_replace('</head>', $preload . "\n" . '</head>', $sCode);
    }
    unset($preloads);
    unset($preload);

    if (strpos($switch, 'images')) {
        $sCode = _convertImagesFromCodeToWebP($sCode, $cfgClient[$client]['path']['frontend'], 0);
    }
    
    # Convert non unix type line breaks to unix type line breaks
    $sCode = str_replace(array("\r\n", "\r"), "\n", $sCode);
    
    if (strpos($switch, 'full') !== false) {
        # Remove preceding and trailing spaces from each line except in <textarea> controls
        $aCnt = explode("\n", $sCode);
        $bTextarea = false;
        for ($i = 0, $n = count($aCnt); $i < $n; $i ++) {
            if (!$bTextarea) {
                $aCnt[$i] = trim($aCnt[$i]);
            }
            if (strpos(strtolower($aCnt[$i]), '<textarea') !== false) {
                $bTextarea = true;
            }
            if (strpos(strtolower($aCnt[$i]), '</textarea>') !== false) {
                $bTextarea = false;
            }
        }
        $sCode = implode("\n", $aCnt);
        unset($aCnt);
    }
    
    # Remove empty conditional comment tags
    $p1 = strpos($sCode, '<!--[if');
    while ($p1 !== false) {
        # First get the exact start comment
        $p2 = strpos($sCode, '>', $p1);
        $c = substr($sCode, $p1, (($p2 - $p1) + 1));
        if (substr($sCode, ($p2 + 1), 5) == '<!-->') {
            $c .= '<!-->';
        }
        $p2 = strpos($sCode, 'endif]-->', $p1);
        $tmp = substr($sCode, $p1, (($p2 - $p1) + 9));
        # Remove line breaks and tabs
        $tmp2 = str_replace(array("\r\n", "\r", "\n", "\t"), '', $tmp);
        # Remove spaces
        while (strpos($tmp2, ' ') !== false) {
            $tmp2 = str_replace(' ', '', $tmp2);
        }
        # Now check the length of the resultung string
        if (strlen($tmp2) <= (strlen($c) + 16)) { # (start tag + end tag)
            # The block can be taken out from it's original position
            $sCode = str_replace($tmp, '', $sCode);
        }
        $p1 = strpos($sCode, '<!--[if', ($p1 + 1));
    }
    
    # Remove blank lines except in <textarea> controls
    $aLines = explode("\n", $sCode);
    $sCode = '';
    $bTextarea = false;
    for ($i = 0, $n = count($aLines); $i < $n; $i ++) {
        if ((strlen(trim($aLines[$i]))) || ($bTextarea)) {
            if (strpos(strtolower($aLines[$i]), '<textarea') !== false) {
                $bTextarea = true;
            }
            $sCode .= $aLines[$i] . "\n";
            if (strpos(strtolower($aLines[$i]), '</textarea>') !== false) {
                $bTextarea = false;
            }
        }
    }
    unset($aLines);
    
    # Do some more cleanup
    unset($tmp);
    unset($tmp2);
    unset($char);
    unset($url);
    unset($p1);
    unset($p2);
    unset($p3);
    unset($p4);
    unset($p5);
    
    $sCode = str_replace(
        array(
            $cfgClient[$client]['path']['htmlpath'] . 'data:',
            $cfgClient[$client]['path']['htmlpath'] . 'cache/less_compiled/data:',
            $cfgClient[$client]['path']['htmlpath'] . 'cache/scss_compiled/data:'
        ),
        'data:',
        $sCode
    );

    if (strlen($sDbgCode)) {
        $sCode = str_replace('</html>', $sDbgCode . '</html>', $sCode);
    }
    
    # Convert the DZIRcms Tag to a html comment
    $sCode = str_replace(array('<DZIRcms>', '</DZIRcms>'), array('<!--', '-->'), $sCode);
    
    # Compress the HTML code?
    if ((strstr($_SERVER['HTTP_ACCEPT_ENCODING'], 'gzip')) && (extension_loaded('zlib')) && (!ini_get("zlib.output_compression"))) {
        header('Content-Encoding: gzip');
        $sCode = gzencode($sCode, 6, FORCE_GZIP);
    }
    
    # Return the new page code
    return $sCode;
}
?>